import io
import re
from collections import defaultdict
from pprint import pp

extra_requirements = """
# FORMAT
# Put your extra requirements here in the following format
#
# package[version_required]: tag1, tag2, ...

ttkwidgets: ttkwidgets
tksheet: tksheet
tkinterweb: tkinterweb
tkintertable: tkintertable
tkcalendar: tkcalendar
AwesomeTkinter: awesometkinter
"""


def get_extra_requires(add_all=True):
    ifile = io.StringIO(extra_requirements)
    with ifile as fp:
        extra_deps = defaultdict(set)
        for k in fp:
            if k.strip() and not k.startswith("#"):
                tags = set()
                if ":" in k:
                    k, v = k.split(":")
                    tags.update(vv.strip() for vv in v.split(","))
                # tags.add(re.split("[<=>]", k)[0])
                for t in tags:
                    extra_deps[t].add(k)
        # add tag `all` at the end
        if add_all:
            extra_deps["all"] = set(vv for v in extra_deps.values() for vv in v)

    return extra_deps


if __name__ == "__main__":
    requires = get_extra_requires()
    for key, dep in requires.items():
        print(f"{key} = {list(dep)}")



# encoding: utf-8
import logging
import tkinter as tk


logger = logging.getLogger(__name__)


def bindings(widget, seq):
    return [x for x in widget.bind(seq).splitlines() if x.strip()]


def _funcid(binding):
    return binding.split()[1][3:]


def remove_binding(widget, seq, index=None, funcid=None):
    b = bindings(widget, seq)

    if index is not None:
        try:
            binding = b[index]
            widget.unbind(seq, _funcid(binding))
            b.remove(binding)
        except IndexError:
            logger.info("Binding #%d not defined.", index)
            return

    elif funcid:
        binding = None
        for x in b:
            if _funcid(x) == funcid:
                binding = x
                b.remove(binding)
                widget.unbind(seq, funcid)
                break
        if not binding:
            logger.info('Binding "%s" not defined.', funcid)
            return
    else:
        raise ValueError("No index or function id defined.")

    for x in b:
        widget.bind(seq, "+" + x, 1)



from ...component.builderobject import BuilderObject
from ...component.plugin_engine import (
    IPluginBase,
    IBuilderLoaderPlugin,
    IDesignerPlugin,
    BuilderLoaderPlugin,
)

from ._private import (
    register_builder,
    register_widget,
    register_property,
    register_custom_property,
    copy_custom_property,
)



import tkinter.ttk as ttk

from pygubu.utils.widget import HideableMixin


class HideableFrame(HideableMixin, ttk.Frame):
    """A frame that can be easily hidden.
    Use hidden property to show or hide the frame.

    myframe.hidden = True
    """

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)



import tkinter as tk
from typing import Mapping
from TKinterModernThemes.WidgetFrame import WidgetFrame, Notebook, PanedWindow
from pygubu.api.v1 import BuilderObject
from pygubu.utils.datatrans import ListDTO
from .codegen import TkmtWidgetCodeMixin


def tkmt_to_tkwidget(widget):
    """Get underline tk widget."""
    if isinstance(widget, tk.Widget):
        return widget
    if isinstance(widget, WidgetFrame):
        return widget.master
    if isinstance(widget, Notebook):
        return widget.notebook
    if isinstance(widget, PanedWindow):
        return widget.panedwindow
    return None


# Groups for ordering buttons in designer palette.
GROUP_ROOT = 0
GROUP_CONTAINER = 10
GROUP_DISPLAY = 30
GROUP_INPUT = 60


class CommandProxy:
    def __init__(self):
        self.command = None

    @property
    def __name__(self):
        return f"CommandProxy({self.command})"

    def __call__(self, *args):
        if self.command is not None:
            self.command(*args)


class TkmtWidgetBO(TkmtWidgetCodeMixin, BuilderObject):
    allow_bindings = False
    layout_required = False
    pos_args = ()
    kw_args = ("row", "col", "padx", "pady", "rowspan", "colspan", "sticky")
    args_to_list = ListDTO([], [])  # To process args property

    def __init__(self, builder, wmeta):
        super().__init__(builder, wmeta)
        self.command_proxies: Mapping[str, CommandProxy] = {}
        self.make_resizable = None

    def realize(self, parent, extra_init_args: dict = None):
        master = parent.get_child_master()
        pbag = self._process_properties(tkmt_to_tkwidget(master))
        kargs = self._get_keyword_args(pbag)
        args = self._get_positional_args(pbag)
        self.widget = self.class_(*args, **kargs)
        return self.widget

    def _process_properties(self, tkmaster: tk.Widget) -> dict:
        defaults = self._get_property_defaults(tkmaster)
        pbag = {}
        for pname in self.properties:
            if pname in self.wmeta.properties:
                pvalue = self.wmeta.properties[pname]
                pbag[pname] = self._process_property_value(pname, pvalue)
            elif pname in defaults:
                pbag[pname] = defaults[pname]
        self._post_process_properties(tkmaster, pbag)
        return pbag

    def _post_process_properties(self, tkmaster: tk.Widget, pbag: dict) -> None:
        pass

    def _get_keyword_args(self, bag: dict) -> dict:
        kargs = {}
        for pname in self.properties:
            if pname in self.kw_args and pname in bag:
                kargs[pname] = bag[pname]
        return kargs

    def _get_positional_args(self, bag: dict) -> list:
        args = []
        for pname in self.pos_args:
            if pname in bag:
                value = bag[pname]
                args.append(value)
        return args

    def _get_property_defaults(self, master: tk.Widget = None) -> dict:
        return {}

    def configure(self, target=None):
        if target is None:
            target = self.widget
        for pname, value in self.wmeta.properties.items():
            if pname not in self.pos_args and pname not in self.kw_args:
                self._set_property(target, pname, value)

    def _set_property(self, target_widget, pname, value):
        if pname == "makeResizable":
            self.make_resizable = value
            return
        super()._set_property(target_widget, pname, value)

    def configure_children(self, target=None):
        if target is None:
            target = self.widget
        if self.make_resizable is not None:
            if self.make_resizable == "all":
                target.makeResizable()
            elif self.make_resizable == "recursive":
                target.makeResizable(recursive=True, onlyFrames=False)
            else:
                target.makeResizable(recursive=False, onlyFrames=True)

    def _process_property_value(self, pname, value):
        if pname in self.command_properties:
            cmd_proxy = self.command_proxies.get(pname, None)
            if cmd_proxy is None:
                cmd_proxy = CommandProxy()
                self.command_proxies[pname] = cmd_proxy
            return cmd_proxy
        if pname in ("row", "col", "rowspan", "colspan"):
            return int(value)
        if pname in ("args", "validatecommandargs", "invalidcommandargs"):
            return self.args_to_list.transform(value)
        if pname in ("disabled", "usecommandlineargs", "useconfigfile"):
            return tk.getboolean(value)
        return super()._process_property_value(pname, value)

    def _connect_command(self, cmd_pname, callback):
        self.command_proxies[cmd_pname].command = callback

    def _code_process_property_value(self, targetid, pname, value: str):
        if pname in ("args", "validatecommandargs", "invalidcommandargs"):
            return str(self.args_to_list.transform(value))
        if pname == "variable":
            # variables can be none in some constructors
            # avoid error when builder trys to create a tk variable.
            if value is None:
                return str(value)
        return super()._code_process_property_value(targetid, pname, value)


class WidgetAsMethodBO(TkmtWidgetBO):
    master_add_method = None

    def realize(self, parent, extra_init_args: dict = None):
        master = parent.get_child_master()
        assert self.master_add_method is not None
        add_method = getattr(master, self.master_add_method)
        pbag = self._process_properties(tkmt_to_tkwidget(master))
        kargs = self._get_keyword_args(pbag)
        args = self._get_positional_args(pbag)
        self.widget = add_method(*args, **kargs)
        return self.widget



# encoding: utf8
import os
import sys
import unittest
import tkinter as tk
import tkinter.ttk as ttk
import tkinter.font as tkfont

import fixpath
import pygubu
import support


class TestBuilderFont(unittest.TestCase):
    def setUp(self):
        support.root_deiconify()
        xmldata = "test_builderfont.ui"
        self.builder = builder = pygubu.Builder()
        builder.add_from_file(xmldata)

        self.font = font = tkfont.Font(
            family="Helvetica", size=32, weight="bold"
        )
        builder.set_font("custom_font", font)

        self.widget = builder.get_object("mainwindow")

    def tearDown(self):
        support.root_withdraw()

    @unittest.skip("Not implemented yet")
    def test_get_font(self):
        font = self.builder.get_font("custom_font")
        self.assertEqual(self.font, font)
        self.widget.destroy()

    @unittest.skip("Not implemented yet")
    def test_get_font_notset(self):
        fname = "other_font"
        font = self.builder.get_font(fname)
        self.assertEqual(fname, font)
        self.widget.destroy()


if __name__ == "__main__":
    unittest.main()



import tkinter as tk
import tkinter.ttk as ttk
from collections import defaultdict
from .themes import STANDARD_THEMES
from .builder import ThemeDefinition, BootstrapThemeBuilder


class Style(ttk.Style):
    """Style class to manage bootstrap themes."""

    bs_names = []
    bs_definitions = {}
    bs_builders = {}
    bs_themes_loaded = False
    bs_configured_roots = defaultdict(list)

    def __init__(self, master=None):
        root = None if master is None else master.winfo_toplevel()
        super().__init__(root)
        self._load_standard_themes()

    def _load_standard_themes(self):
        if not self.bs_themes_loaded:
            for name, definition in STANDARD_THEMES.items():
                self.register_theme(
                    ThemeDefinition(
                        name=name,
                        theme_type=definition["type"],
                        colors=definition["colors"],
                    )
                )
            type(self).bs_themes_loaded = True

    @classmethod
    def register_theme(cls, definition: ThemeDefinition):
        cls.bs_names.append(definition.name)
        cls.bs_definitions[definition.name] = definition

    def theme_names(self, bs_themes_only=False):
        if bs_themes_only:
            return tuple(self.bs_names)
        return super().theme_names() + tuple(self.bs_names)

    def theme_use(self, themename=None):
        if themename is None:
            return super().theme_use()
        if themename not in self.bs_names:
            # raise ValueError(f"Bootstrap theme '{themename}' not found.")
            return super().theme_use(themename)
        current_theme = super().theme_use()
        if current_theme == themename:
            # Already using theme: themename
            # Do nothing
            return
        builder = type(self).bs_builders.get(themename, None)
        if builder is None:
            builder = BootstrapThemeBuilder(self.bs_definitions[themename])
            type(self).bs_builders[themename] = builder
        if themename not in self.bs_configured_roots[self.master]:
            builder.create(self.master)
            builder.apply(self.master)
            self.bs_configured_roots[self.master].append(themename)
        else:
            builder.apply(self.master)
        builder.update_current_widgets(self.master)

    @classmethod
    def get_generated_styles(cls):
        return list(BootstrapThemeBuilder.generated_styles)



# encoding: utf-8
import os
import sys
import unittest
import tkinter as tk
import tkinter.ttk as ttk

import fixpath
import pygubu
import support


class MyContainer(object):
    pass


class TestText(unittest.TestCase):
    def setUp(self):
        support.root_deiconify()
        xmldata = "test_import_variables.ui"
        self.builder = builder = pygubu.Builder()
        builder.add_from_file(xmldata)
        self.widget = builder.get_object("mainwindow")

    def tearDown(self):
        support.root_withdraw()

    def test_import_variables(self):
        container = MyContainer()
        self.builder.import_variables(container)
        self.assertIsInstance(container.myvar_string, tk.StringVar)
        self.assertIsInstance(container.myvar_int, tk.IntVar)
        self.assertIsInstance(container.myvar_double, tk.DoubleVar)
        self.assertIsInstance(container.myvar_boolean, tk.BooleanVar)



import logging
import sys
import tkinter as tk
from pathlib import Path
from collections import namedtuple
from dataclasses import dataclass
from enum import Enum
from typing import Any
from .exceptions import ImageFormatNotSupportedError, ImageNotFoundError
from .config import (
    TK_IMAGE_FORMATS,
    TK_BITMAP_FORMATS,
    BITMAP_TEMPLATE,
    TK_PHOTO_FORMATS,
)


if sys.version_info < (3, 9):
    import importlib_resources as resources
else:
    import importlib.resources as resources


logger = logging.getLogger(__name__)


def _iter_package_files(pkg: str):
    subpkg = []
    try:
        for r in resources.files(pkg).iterdir():
            if r.is_file():
                yield r
            if r.is_dir():
                subpkg.append(r.name)
    except NotADirectoryError:
        pass
    for s in subpkg:
        yield from _iter_package_files(f"{pkg}/{s}")


class PitType(Enum):
    PATH = 1
    PACKAGE = 2


@dataclass
class StockItem:
    def create_image(self, *, tk_master=None, custom_loader=None):
        ...


@dataclass
class ImgFromPath(StockItem):
    fpath: Any = None

    def create_image(self, *, tk_master=None, custom_loader=None):
        file_ext = self.fpath.suffix.lower()
        if custom_loader is not None:
            img = custom_loader(PitType.PATH, self.fpath, tk_master)
        elif file_ext in TK_PHOTO_FORMATS:
            img = tk.PhotoImage(file=self.fpath, master=tk_master)
        elif file_ext in TK_BITMAP_FORMATS:
            img = tk.BitmapImage(file=self.fpath, master=tk_master)
        else:
            img = self._create_with_pillow(self.fpath, tk_master)
        return img

    def _create_with_pillow(self, fpath, tk_master):
        try:
            from PIL import Image, ImageTk

            aux = Image.open(fpath)
            img = ImageTk.PhotoImage(aux, master=tk_master)
            return img
        except ModuleNotFoundError:
            msg = f"Error loading {self.fpath}, image format not supported."
            raise ImageFormatNotSupportedError(msg)


@dataclass
class ImgFromPackage(ImgFromPath):
    def create_image(self, *, tk_master=None, custom_loader=None):
        file_ext = Path(str(self.fpath)).suffix.lower()

        with resources.as_file(self.fpath) as file:
            if custom_loader is not None:
                img = custom_loader(PitType.PACKAGE, file, tk_master)
            elif file_ext in TK_PHOTO_FORMATS:
                img = tk.PhotoImage(file=file, master=tk_master)
            elif file_ext in TK_BITMAP_FORMATS:
                img = tk.BitmapImage(file=file, master=tk_master)
            else:
                img = self._create_with_pillow(file, tk_master)
        return img


@dataclass
class ImgFromData(StockItem):
    data: Any = None
    format: str = None

    def create_image(self, *, tk_master=None, custom_loader=None):
        return tk.PhotoImage(
            format=self.format, data=self.data, master=tk_master
        )


class StockRegistry:
    """Maintain image source definitions to load."""

    def __init__(self):
        self._stock = {}
        self._formats = TK_IMAGE_FORMATS
        self._resource_pit = {}

    def _register_stock_item(self, image_id, cache_item):
        if image_id in self._stock:
            logger.warning("Warning, replacing resource %s", image_id)
        self._stock[image_id] = cache_item

    def register(self, image_id, filename):
        """Register a image file using image_id"""
        fpath = Path(filename) if isinstance(filename, str) else filename
        self._register_stock_item(image_id, ImgFromPath(fpath))
        logger.info("%s registered as %s", filename, image_id)

    def register_from_package(self, image_id, fpath):
        self._register_stock_item(image_id, ImgFromPackage(fpath))
        logger.info("%s registered as %s", fpath, image_id)

    def register_from_data(self, image_id, format, data):
        """Register a image data using image_id"""

        if image_id in self._stock:
            logger.warning("Warning, replacing resource %s", image_id)
        self._stock[image_id] = ImgFromData(data, format)
        logger.info("%s registered as %s", "data", image_id)

    def is_registered(self, image_id):
        return image_id in self._stock

    def get_item(self, image_id):
        return self._stock[image_id]

    def register_all_from_dir(
        self, dir_path, prefix=None, ext=None, recurse=False, fullname_key=False
    ):
        """List files from dir_path and register images with
            filename as key (without extension)

        :param str dir_path: path to search for images.
        :param str prefix: Additionaly a prefix for the key can be provided,
            so the resulting key will be prefix + filename
        :param iterable ext: list of file extensions to load. Defaults to
            tk supported image extensions. Example ('.jpg', '.png')
        :param boolean recurse: search recursivelly.
        :param boolean fullname_key: use filename and extension as key
        """
        prefix = "" if prefix is None else prefix
        if ext is None:
            ext = TK_IMAGE_FORMATS

        path_gen = Path(dir_path).iterdir()
        if recurse:
            path_gen = Path(dir_path).glob("**/*")

        for filename in path_gen:
            if filename.is_file():
                name = filename.stem
                file_ext = filename.suffix
                if file_ext in ext:
                    fkey = f"{prefix}{name}"
                    if fullname_key:
                        fkey = f"{prefix}{name}{file_ext}"
                    self.register(fkey, filename)

    def register_all_from_pkg(
        self, pkg, prefix=None, ext=None, recurse=False, fullname_key=False
    ):
        """List files from package and register images with
            filename as key (without extension)

        :param str pkg: package to search for images.
        :param str prefix: Additionaly a prefix for the key can be provided,
            so the resulting key will be prefix + filename
        :param iterable ext: list of file extensions to load. Defaults to
            tk supported image extensions. Example ('.jpg', '.png')
        :param boolean recurse: search recursivelly.
        :param boolean fullname_key: use filename and extension as key
        """
        prefix = "" if prefix is None else prefix
        if ext is None:
            ext = TK_IMAGE_FORMATS
        path_gen = resources.files(pkg).iterdir()
        if recurse:
            path_gen = _iter_package_files(pkg)
        for pkg_path in path_gen:
            if pkg_path.is_file():
                fpath = Path(str(pkg_path))
                name = fpath.stem
                file_ext = fpath.suffix
                if file_ext in ext:
                    image_id = f"{prefix}{name}"
                    if fullname_key:
                        image_id = f"{prefix}{name}{file_ext}"
                    self.register_from_package(image_id, pkg_path)

    def add_resource_path(self, path):
        if path not in self._resource_pit:
            self._resource_pit[path] = PitType.PATH

    def add_resource_package(self, package):
        if package not in self._resource_pit:
            self._resource_pit[package] = PitType.PACKAGE

    def find_and_register(self, image_id):
        """Find and register image from the resource pits."""
        image_path = None
        pattern = f"*{image_id}"
        for pit, pit_type in self._resource_pit.items():
            if pit_type == PitType.PATH:
                try:
                    image_path = self._find_in_path(pit, pattern)
                except TypeError:
                    pass
                if image_path:
                    self.register(image_id, image_path)
                    break
            else:
                try:
                    image_path = self._find_in_package(pit, pattern)
                except ModuleNotFoundError:
                    pass
                if image_path:
                    self.register_from_package(image_id, image_path)
                    break
        if image_path is None:
            msg = f"Error: image {image_id} not found in resource pits."
            raise ImageNotFoundError(msg)

    @staticmethod
    def _find_in_path(path_src, pattern):
        found = None
        for p in Path(path_src).glob("**/*"):
            if p.is_file() and p.match(pattern):
                found = p
                break
        return found

    @staticmethod
    def _find_in_package(pkg_src, pattern):
        found = None
        for r in _iter_package_files(pkg_src):
            p2 = Path(str(r))
            if r.is_file() and p2.match(pattern):
                found = r
                break
        return found

    def as_iconbitmap(self, image_id):
        """Get image path for use in iconbitmap property"""
        img = None
        if image_id in self._stock:
            cache_item = self._stock[image_id]
            if isinstance(cache_item, ImgFromPath):
                fpath = cache_item.fpath
                file_ext = fpath.suffix.lower()

                if file_ext in TK_BITMAP_FORMATS:
                    img = BITMAP_TEMPLATE.format(fpath)
        return img



hiddenimports = [
    "pygubu.plugins.pygubu.accordionframe_bo",
    "pygubu.plugins.pygubu.calendarframe_bo",
    "pygubu.plugins.pygubu.colorinput_bo",
    "pygubu.plugins.pygubu.combobox_bo",
    "pygubu.plugins.pygubu.dialog_bo",
    "pygubu.plugins.pygubu.dockfw_bo",
    "pygubu.plugins.pygubu.editabletreeview_bo",
    "pygubu.plugins.pygubu.filterabletreeview_bo",
    "pygubu.plugins.pygubu.flodgauge_bo",
    "pygubu.plugins.pygubu.forms",
    "pygubu.plugins.pygubu.forms.pygubuwidget_bo",
    "pygubu.plugins.pygubu.forms.tkwidget_bo",
    "pygubu.plugins.pygubu.forms.ttkwidget_bo",
    "pygubu.plugins.pygubu.hideableframe_bo",
    "pygubu.plugins.pygubu.pathchooserinput_bo",
    "pygubu.plugins.pygubu.scrollbarhelper_bo",
    "pygubu.plugins.pygubu.scrolledframe_bo",
    "pygubu.plugins.pygubu.tkscrollbarhelper_bo",
    "pygubu.plugins.pygubu.tkscrolledframe_bo",
    "pygubu.plugins.tk.scrolledtext_bo",
    "pygubu.plugins.tk.tkstdwidgets",
    "pygubu.plugins.ttk.ttkstdwidgets",
]



import tkinter as tk
import tkinter.ttk as ttk


class EntryWPlaceholder(ttk.Entry):
    _PHKEY = "placeholder"

    def __init__(self, master=None, **kw):
        self.placeholder = kw.pop(self._PHKEY, "")
        super().__init__(master, **kw)

        self.bind("<FocusIn>", self._on_focusin)
        self.bind("<FocusOut>", self._on_focusout)

        if "textvariable" in kw:
            var = kw["textvariable"]
            var.trace_add("write", self._trace_var)
        self._put_placeholder()

    def _trace_var(self, var, *args):
        print(f"tracing {var}", self.__get())
        if self.focus_displayof() is not self:
            print("putting placeholder after var trace")
            self._put_placeholder(self.__get())

    def _put_placeholder(self):
        print("on put_placeholder", self.__get(), not self.__get())
        if not self.__get():
            print("inserting placeholder")
            self.insert(0, self.placeholder)

    def _on_focusin(self, event):
        print("on_focusin")
        if self.__get() == self.placeholder:
            print("deleting placeholder")
            self.delete("0", "end")

    def _on_focusout(self, event):
        self._put_placeholder()

    def __get(self):
        return super().get()

    def get(self):
        value = super().get()
        if value and (value == self.placeholder):
            value = ""
        return value

    def configure(self, cnf=None, **kw):
        key = self._PHKEY
        if cnf:
            if cnf == key:
                return (key, self.cget(key))
            return super().configure(cnf, **kw)
        if key in kw:
            self.placeholder = kw.pop(key)
            self._put_placeholder()
        return super().configure(cnf, **kw)

    config = configure

    def cget(self, key):
        if key == self._PHKEY:
            return self.placeholder
        return super().cget(key)

    __getitem__ = cget


if __name__ == "__main__":
    root = tk.Tk()
    var = tk.StringVar()
    username = EntryWPlaceholder(root, textvariable=var)
    username["placeholder"] = "--USERNAME--"

    var2 = tk.StringVar()
    password = EntryWPlaceholder(root, textvariable=var2)
    password.configure(style="MyPasswordEntry.TEntry")
    password.configure(placeholder="--PASSWORD--")

    username.pack()
    password.pack()

    def on_click():
        print(f"user: {username.get()}")
        print(f"pass: {password.get()}")

    def on_start():
        var.set("username")

    btn = ttk.Button(root, text="Test", takefocus=True, command=on_click)
    btn.pack()
    root.after(800, on_start)
    root.mainloop()



import math
import time
import tkinter as tk


class TkPhotoDraw:
    """A fallback class for drawing in a tk.PhotoImage."""

    def __init__(self, master):
        self.tk_master: tk.Widget = master
        self.canvas: tk.PhotoImage = None

    def _tk_call(self, *args, **kw):
        return self.tk_master.tk.call(*args, **kw)

    # Canvas related functions begin
    def canvas_new(self, **photo_args):
        if "master" not in photo_args:
            photo_args["master"] = self.tk_master
        self.canvas = tk.PhotoImage(**photo_args)
        return self.canvas

    def canvas_blank(self):
        self.canvas.blank()

    def canvas_scale(self, xf, yf=0) -> tk.PhotoImage:
        """Scale canvas to new image.
        Returns:
           tk.PhotoImage
        """
        mode = "-subsample"
        if abs(xf) < 1:
            xf = 1.0 / xf
            if abs(yf) < 1 and yf != 0:
                yf = 1.0 / yf
        elif xf >= 0 and yf >= 0:
            mode = "-zoom"
        if yf == 0:
            yf = xf
        xf, yf = int(xf), int(yf)
        tmp = tk.PhotoImage(master=self.tk_master)
        self.tk_master.tk.call(
            tmp.name, "copy", self.canvas.name, "-shrink", mode, xf, yf
        )
        return tmp

    def canvas_rotate(self, angle: int, bg_color=None):
        if bg_color is None:
            bg_color = "#000000"
        w = self.canvas.width()
        h = self.canvas.height()
        tmp = self.canvas.copy()
        self.canvas.blank()
        if angle == 180 or angle == -180:
            self._tk_call(
                self.canvas.name, "copy", tmp.name, "-subsample", -1, -1
            )
        elif angle == 90:
            self._rotate_90(bg_color, w, h, tmp)
        elif angle == -90 or angle == 270:
            self._rotate_270(bg_color, w, h, tmp)
        else:
            self._handle_other_rotations(angle, bg_color, w, h, tmp)
        del tmp

    def _handle_other_rotations(self, angle, bg_color, w, h, tmp):
        buf = []
        alpha = []
        bg_rgb = self.tk_master.winfo_rgb(bg_color)
        a = math.atan(1) * 8 * angle / 360
        xm = w / 2
        ym = h / 2
        w2 = round(abs(w * math.cos(a)) + abs(h * math.sin(a)))
        xm2 = w2 / 2
        h2 = round(abs(h * math.cos(a)) + abs(w * math.sin(a)))
        ym2 = h2 / 2
        self.canvas.config(width=w2, height=h2)
        for i in range(0, h2):
            to_x = -1
            for j in range(0, w2):
                rad = math.hypot(ym2 - i, xm2 - j)
                th = math.atan2(ym2 - i, xm2 - j) + a
                x = xm - rad * math.cos(th)
                if x < 0 or x >= w:
                    alpha.append((i, j))
                    continue
                y = ym - rad * math.sin(th)
                if y < 0 or y >= h:
                    alpha.append((i, j))
                    continue
                x0 = int(x)
                x1 = x0 + 1 if (x0 + 1) < w else x0
                dx = x1 - x
                y0 = int(y)
                y1 = y0 + 1 if (y0 + 1) < h else y0
                dy = y1 - y
                R = G = B = 0

                rgb1 = self._get_img_pixel(tmp, x0, y0)
                rgb2 = self._get_img_pixel(tmp, x0, y1)
                rgb3 = self._get_img_pixel(tmp, x1, y0)
                rgb4 = self._get_img_pixel(tmp, x1, y1)

                if (
                        rgb1 == bg_rgb
                        or rgb2 == bg_rgb
                        or rgb3 == bg_rgb
                        or rgb4 == bg_rgb
                    ):
                    alpha.append((i, j))
                else:
                    r, g, b = rgb1
                    R = R + r * dx * dy
                    G = G + g * dx * dy
                    B = B + b * dx * dy
                    r, g, b = rgb2
                    R = R + r * dx * (1 - dy)
                    G = G + g * dx * (1 - dy)
                    B = B + b * dx * (1 - dy)
                    r, g, b = rgb3
                    R = R + r * (1 - dx) * dy
                    G = G + g * (1 - dx) * dy
                    B = B + b * (1 - dx) * dy
                    r, g, b = rgb4
                    R = R + r * (1 - dx) * (1 - dy)
                    G = G + g * (1 - dx) * (1 - dy)
                    B = B + b * (1 - dx) * (1 - dy)
                rgb = (round(R), round(G), round(B))
                buf.append("#{:02x}{:02x}{:02x}".format(*rgb))
                if to_x == -1:
                    to_x = j
            if to_x >= 0:
                self.canvas.put(buf, to=(i, to_x))
                for x, y in alpha:
                    self.canvas.transparency_set(x, y, True)
                buf.clear()
                alpha.clear()
        del buf

    def _rotate_270(self, bg_color, w, h, tmp):
        bg_rgb = self.tk_master.winfo_rgb(bg_color)
        buf = []
        alpha = []
        for i in range(0, w):
            row = []
            for j in range(h - 1, -1, -1):
                rgb = self._get_img_pixel(tmp, i, j)
                if rgb == bg_rgb:
                    alpha.append((i, j))
                row.append("#{:02x}{:02x}{:02x}".format(*rgb))
            buf.append(row)
        self.canvas.config(width=h, height=w)
        self.canvas.put(buf)
        for x, y in alpha:
            self.canvas.transparency_set(y, x, True)
        del alpha
        del buf

    def _get_img_pixel(self, img, x:int, y:int):
        rgb = img.get(x, y)
        if isinstance(rgb, str):
            # Fix python tk 8.5 issue
            r, g, b = rgb.split()
            rgb = (int(r), int(g), int(b))
        return rgb

    def _rotate_90(self, bg_color, w, h, tmp):
        bg_rgb = self.tk_master.winfo_rgb(bg_color)
        buf = []
        alpha = []
        for i in range(w - 1, -1, -1):
            row = []
            for j in range(0, h):
                rgb = self._get_img_pixel(tmp, i, j)
                if rgb == bg_rgb:
                    alpha.append((i, j))
                row.append("#{:02x}{:02x}{:02x}".format(*rgb))
            buf.append(row)
        self.canvas.config(width=h, height=w)
        self.canvas.put(buf)
        for x, y in alpha:
            self.canvas.transparency_set(y, x, True)
        del alpha
        del buf

    # Canvas related functions end

    @staticmethod
    def bresenham_line(x1, y1, x2, y2):
        dx = abs(x2 - x1)
        dy = abs(y2 - y1)
        slope = dy > dx
        if slope:
            x1, y1 = y1, x1
            x2, y2 = y2, x2
        if x1 > x2:
            x1, x2 = x2, x1
            y1, y2 = y2, y1
        dx = abs(x2 - x1)
        dy = abs(y2 - y1)
        error = dx // 2
        y = y1
        ystep = 1 if y1 < y2 else -1
        for x in range(x1, x2 + 1):
            coord = (y, x) if slope else (x, y)
            yield coord
            error -= dy
            if error < 0:
                y += ystep
                error += dx

    def _line(self, x1, y1, x2, y2, color):
        """Draw line."""
        for x, y in Draw.bresenham_line(x1, y1, x2, y2):
            self.canvas.put(color, to=(x, y))

    def _draw_rect_filled(self, x1, y1, x2, y2, color):
        # self.canvas.put(color, to=(x1, y1, x2 + 1, y2 + 1))
        self.canvas.put(color, to=(x1, y1, x2, y2))

    def _get_ink(self, fill, outline):
        fill_ink = fill
        border_ink = outline

        if fill is None and outline is None:
            fill_ink = "black"
            border_ink = "black"
        elif fill is None:
            fill_ink = outline
        elif outline is None:
            border_ink = fill_ink
        return fill_ink, border_ink

    def rectangle(self, x1, y1, x2, y2, *, fill=None, outline=None, stroke=1):
        fill_ink, border_ink = self._get_ink(fill, outline)

        if fill_ink == border_ink and fill:
            self._draw_rect_filled(x1, y1, x2, y2, fill_ink)
        else:
            # top line
            self._draw_rect_filled(x1, y1, x2, y1 + stroke, border_ink)
            # bottom
            self._draw_rect_filled(x1, y2 - stroke, x2, y2, border_ink)
            # left
            self._draw_rect_filled(x1, y1, x1 + stroke, y2, border_ink)
            # right
            self._draw_rect_filled(x2 - stroke, y1, x2, y2, border_ink)
            # interior
            if fill:
                self._draw_rect_filled(
                    x1 + stroke, y1 + stroke, x2 - stroke, y2 - stroke, fill_ink
                )

    def rounded_rectangle(
        self, x1, y1, x2, y2, *, fill=None, outline=None, stroke=1, radio=0
    ):
        fill_ink, border_ink = self._get_ink(fill, outline)

        # left border
        self._draw_rect_filled(
            x1, y1 + radio, x1 + stroke, y2 - radio, border_ink
        )
        # right border
        self._draw_rect_filled(
            x2 - stroke, y1 + radio, x2, y2 - radio, border_ink
        )
        # top border
        self._draw_rect_filled(
            x1 + radio, y1, x2 - radio, y1 + stroke, border_ink
        )
        # bottom border
        self._draw_rect_filled(
            x1 + radio, y2 - stroke, x2 - radio, y2, border_ink
        )

        if fill:
            # center rectangle
            self._draw_rect_filled(
                x1 + stroke, y1 + radio, x2 - stroke, y2 - radio, fill_ink
            )
            # top
            self._draw_rect_filled(
                x1 + radio, y1 + stroke, x2 - radio, y1 + radio, fill_ink
            )
            # bottom
            self._draw_rect_filled(
                x1 + radio, y2 - radio, x2 - radio, y2 - stroke, fill_ink
            )

        ix1, iy1, ix2, iy2 = (  # noqa: F841
            x1 + stroke,
            y1 + stroke,
            x2 - stroke,
            y2 - stroke,
        )
        diam = radio * 2
        cx1, cy1, cx2, cy2 = (x1, y1, x2, y2)
        cboxes = (
            (cx1, cy1, cx1 + diam, cy1 + diam, 0),
            (cx2 - diam, cy1, cx2, cy1 + diam, 1),
            (cx1, cy2 - diam, cx1 + diam, cy2, 2),
            (cx2 - diam, cy2 - diam, x2, cy2, 3),
        )
        for cx1, cy1, cx2, cy2, cuadrant in cboxes:
            self.circle_cuadrant(
                cx1,
                cy1,
                cx2,
                cy2,
                cuadrant,
                outline=outline,
                fill=fill,
                stroke=stroke,
            )

    def circle(self, x1, y1, x2, y2, *, fill=None, outline=None, stroke=1):
        d = x2 - x1
        r = int(d / 2)
        # center
        x = x1 + r
        y = y1 + r

        fill_ink, border_ink = self._get_ink(fill, outline)

        # will add +1 to y2 because I'm drawing using
        # filled rectangles from tk photo image

        if fill_ink == border_ink and fill:
            self._draw_rect_filled(x - r, y, x + r, y + 1, fill_ink)
            for i in range(1, r):
                a = int(math.sqrt(r * r - i * i))  # Pythagoras
                self._draw_rect_filled(x - a, y - i, x + a, y - i + 1, fill_ink)
                self._draw_rect_filled(x - a, y + i, x + a, y + i + 1, fill_ink)
        else:
            ix1 = x1 + stroke
            ix2 = x2 - stroke
            dd = ix2 - ix1
            ir = int(dd / 2)
            ix = ix1 + ir
            iy = y1 + stroke + ir

            for i in range(0, r):
                a = r if i == 0 else int(math.sqrt(r * r - i * i))  # Pythagoras
                if i > ir:
                    self._draw_rect_filled(
                        x - a, y - i, x + a, y - i + 1, border_ink
                    )
                    self._draw_rect_filled(
                        x - a, y + i, x + a, y + i + 1, border_ink
                    )
                else:
                    ia = (
                        ir if i == 0 else int(math.sqrt(ir * ir - i * i))
                    )  # Pythagoras
                    x1, y1, x2, y2 = (x - a, y - i, x + a, y - i)
                    ix1, iy1, ix2, iy2 = (ix - ia, iy - i, ix + ia, iy - i)
                    self._draw_rect_filled(x1, y1, ix1, iy1 + 1, border_ink)
                    self._draw_rect_filled(ix2, iy2, x2, y2 + 1, border_ink)
                    if fill:
                        self._draw_rect_filled(ix1, iy1, ix2, iy2 + 1, fill_ink)

                    x1, y1, x2, y2 = (x - a, y + i, x + a, y + i)
                    ix1, iy1, ix2, iy2 = (ix - ia, iy + i, ix + ia, iy + i)
                    self._draw_rect_filled(x1, y1, ix1, iy1 + 1, border_ink)
                    self._draw_rect_filled(ix2, iy2, x2, y2 + 1, border_ink)
                    if fill:
                        self._draw_rect_filled(ix1, iy1, ix2, iy2 + 1, fill_ink)

    def circle_cuadrant(
        self, x1, y1, x2, y2, cuadrant, *, fill=None, outline=None, stroke=1
    ):
        """Draw a circle cuadrant
        Circle cuadrants for this function:
        0 - 1
        2 - 3
        """
        d = x2 - x1
        r = int(d / 2)
        # center
        x = x1 + r
        y = y1 + r

        fill_ink, border_ink = self._get_ink(fill, outline)

        # will add +1 to x2 and y2 because I'm drawing using
        # filled rectangles from tk photo image
        ix1 = x1 + stroke
        ix2 = x2 - stroke
        dd = ix2 - ix1
        ir = int(dd / 2)
        ix = ix1 + ir
        iy = y1 + stroke + ir

        for i in range(0, r):
            a = r if i == 0 else int(math.sqrt(r * r - i * i))  # Pythagoras
            if i > ir:
                # UPPER BORDERS
                if cuadrant == 0:
                    self._draw_rect_filled(
                        x - a, y - i, x, y - i + 1, border_ink
                    )
                    pass
                if cuadrant == 1:
                    self._draw_rect_filled(
                        x, y - i, x + a, y - i + 1, border_ink
                    )
                # BOTTOM BORDERS
                if cuadrant == 2:
                    self._draw_rect_filled(
                        x - a, y + i, x, y + i + 1, border_ink
                    )
                if cuadrant == 3:
                    self._draw_rect_filled(
                        x, y + i, x + a, y + i + 1, border_ink
                    )
            else:
                self._handle_lower_i(cuadrant, fill, x, y, fill_ink, border_ink, ir, ix, iy, i, a)

    def _handle_lower_i(self, cuadrant, fill, x, y, fill_ink, border_ink, ir, ix, iy, i, a):
        ia = (
                    ir if i == 0 else int(math.sqrt(ir * ir - i * i))
                )  # Pythagoras
        x1, y1, x2, y2 = (x - a, y - i, x + a, y - i)
        ix1, iy1, ix2, iy2 = (ix - ia, iy - i, ix + ia, iy - i)
                # upper left
        if 0 == cuadrant:
            self._draw_rect_filled(x1, y1, ix1, iy1 + 1, border_ink)
            if fill:
                self._draw_rect_filled(ix1, iy1, x, iy1 + i, fill_ink)
                # upper right
        if 1 == cuadrant:
            self._draw_rect_filled(ix2, iy2, x2, y2 + 1, border_ink)
            if fill:
                self._draw_rect_filled(x, iy1, ix2, iy2 + 1, fill_ink)

        x1, y1, x2, y2 = (x - a, y + i, x + a, y + i)
        ix1, iy1, ix2, iy2 = (ix - ia, iy + i, ix + ia, iy + i)
                # bottom left
        if 2 == cuadrant:
            self._draw_rect_filled(x1, y1, ix1, iy1 + 1, border_ink)
            if fill:
                self._draw_rect_filled(ix1, iy1, x, iy2 + 1, fill_ink)
                # bottom right
        if 3 == cuadrant:
            self._draw_rect_filled(ix2, iy2, x2, y2 + 1, border_ink)
            if fill:
                self._draw_rect_filled(x, iy1, ix2, iy2 + 1, fill_ink)

    def triangle(self, x1, y1, x2, y2, x3, y3, color):
        self._line(x1, y1, x2, y2, color)
        self._line(x2, y2, x3, y3, color)
        self._line(x3, y3, x1, y1, color)

    def triangle_filled(self, x1, y1, x2, y2, x3, y3, color):
        x1, y1, x2, y2, x3, y3 = self._sort_verticesby_y(x1, y1, x2, y2, x3, y3)
        if y2 == y3:
            self._fill_bottom_flat_tri(x1, y1, x2, y2, x3, y3, color)
        elif y1 == y2:
            self._fill_top_flat_tri(x1, y1, x2, y2, x3, y3, color)
        else:
            newx = int(x1 + ((y2 - y1) / (y3 - y1)) * (x3 - x1))
            newy = y2
            self._fill_bottom_flat_tri(x1, y1, x2, y2, newx, newy, color)
            self._fill_top_flat_tri(x2, y2, newx, newy, x3, y3, color)

    def _fill_bottom_flat_tri(self, x1, y1, x2, y2, x3, y3, color):
        slope1 = (x2 - x1) / (y2 - y1)
        slope2 = (x3 - x1) / (y3 - y1)
        hx1 = x1
        hx2 = x1 + 0.5
        for scanline_y in range(y1, y2):
            self._draw_rect_filled(
                int(hx1), scanline_y, int(hx2), scanline_y + 1, color
            )
            hx1 += slope1
            hx2 += slope2

    def _fill_top_flat_tri(self, x1, y1, x2, y2, x3, y3, color):
        slope1 = (x3 - x1) / (y3 - y1)
        slope2 = (x3 - x2) / (y3 - y2)
        hx1 = x3
        hx2 = x3 + 0.5
        for scanline_y in range(y3, y1 - 1, -1):
            self._draw_rect_filled(
                int(hx1), scanline_y, int(hx2), scanline_y + 1, color
            )
            hx1 -= slope1
            hx2 -= slope2

    def _sort_verticesby_y(self, x1, y1, x2, y2, x3, y3):
        verts = sorted([(x1, y1), (x2, y2), (x3, y3)], key=lambda x: x[1])
        return (p for v in verts for p in v)

    def convex_poly(self, points, color):
        if self.is_convex(points):
            # points = sorted(points, key=lambda x: x[1])
            for t in self._tripoly(points):
                self.triangle_filled(*t[0], *t[1], *t[2], color)
        else:
            raise Exception("Not a convex polygon.")

    def _cross_prod(self, points):
        # A[0][0]  px1 [0][1]  py1
        # A[1][0]  px2 A[1][1]  py2
        # A[2][0]  px3 A[2][1]  py3

        x1 = points[1][0] - points[0][0]
        y1 = points[1][1] - points[0][1]
        x2 = points[2][0] - points[0][0]
        y2 = points[2][1] - points[0][1]
        # Return cross product
        return x1 * y2 - y1 * x2

    def _tripoly(self, poly):
        return [(poly[0], b, c) for b, c in zip(poly[1:], poly[2:])]

    def is_convex(self, points):
        # Stores count of
        # edges in polygon
        N = len(points)
        # Stores direction of cross product
        # of previous traversed edges
        prev = 0
        # Stores direction of cross product
        # of current traversed edges
        curr = 0
        # Traverse the array
        for i in range(N):
            # Stores three adjacent edges
            # of the polygon
            temp = (points[i], points[(i + 1) % N], points[(i + 2) % N])
            # Update curr
            curr = self._cross_prod(temp)
            # If curr is not equal to 0
            if curr != 0:
                # If direction of cross product of
                # all adjacent edges are not same
                if curr * prev < 0:
                    return False
                else:
                    # Update curr
                    prev = curr

        return True


Draw = TkPhotoDraw


if __name__ == "__main__":
    root = tk.Tk()
    root.imgs = []

    draw = Draw(root)
    # rectangle
    rr = draw.canvas_new(width=300, height=100)
    root.imgs.append(rr)
    draw.rectangle(10, 10, 90, 90, outline="green", stroke=1)
    draw.rectangle(110, 10, 190, 90, fill="lightgreen")
    draw.rectangle(
        210, 10, 290, 90, fill="lightgreen", outline="green", stroke=10
    )
    label = tk.Label(root, image=rr)
    label.pack()

    # rounded rectangle
    rr = draw.canvas_new(width=300, height=100)
    root.imgs.append(rr)
    draw.rounded_rectangle(10, 10, 90, 90, radio=20, outline="green", stroke=1)
    draw.rounded_rectangle(110, 10, 190, 90, radio=20, fill="lightgreen")
    draw.rounded_rectangle(
        210,
        10,
        290,
        90,
        radio=20,
        fill="lightgreen",
        outline="green",
        stroke=10,
    )
    label = tk.Label(root, image=rr)
    label.pack()

    # circle cuadrant
    rr = draw.canvas_new(width=300, height=100)
    root.imgs.append(rr)
    draw.circle_cuadrant(10, 10, 90, 90, 3, outline="green")
    draw.circle_cuadrant(110, 10, 190, 90, 3, fill="lightgreen")
    draw.circle_cuadrant(
        210, 10, 290, 90, 3, fill="lightgreen", outline="green", stroke=10
    )
    label = tk.Label(root, image=rr)
    label.pack()

    # circle
    rr = draw.canvas_new(width=300, height=100)
    root.imgs.append(rr)
    draw.circle(10, 10, 90, 90, outline="green")
    draw.circle(110, 10, 190, 90, fill="lightgreen")
    draw.circle(210, 10, 290, 90, fill="lightgreen", outline="green", stroke=10)
    label = tk.Label(root, image=rr)
    label.pack()

    # triangle
    rr = draw.canvas_new(width=300, height=100)
    root.imgs.append(rr)
    draw.triangle(10, 90, 50, 0, 90, 90, "green")
    # triangle filled
    draw.triangle_filled(110, 90, 150, 0, 190, 90, "green")
    draw.triangle_filled(210, 0, 290, 0, 250, 90, "green")
    label = tk.Label(root, image=rr)
    label.pack()

    # shape
    rr = draw.canvas_new(width=134, height=134)
    root.imgs.append(rr)
    shape = [
        (34, 76, 38, 72, 43, 70),
        (35, 77, 38, 88, 43, 98),
        (34, 76, 52, 82, 44, 100),
        (34, 76, 51, 82, 44, 70),
        (43, 100, 51, 82, 53, 98),
        (51, 82, 53, 98, 64, 82),
        (52, 81, 64, 82, 85, 53),
        (52, 81, 85, 53, 71, 53),
        (85, 53, 71, 53, 100, 34),
        (71, 53, 100, 34, 85, 35),
        (100, 34, 85, 35, 96, 30),
        (85, 35, 90, 31, 97, 30),
        (96, 30, 100, 34, 101, 30),
    ]
    for x1, y1, x2, y2, x3, y3 in shape:
        draw.triangle_filled(x1, y1, x2, y2, x3, y3, "brown")
    label = tk.Label(root, image=rr)
    label.pack()

    points = [[0, 0], [0, 10], [10, 10], [10, 0]]
    points2 = [(0, 0), (0, 10), (5, 5), (10, 10), (10, 0)]

    for p in (points, points2):
        if draw.is_convex(p):
            print("Yes")
            draw.convex_poly(p, "red")
        else:
            print("No")

    root.mainloop()



import awesometkinter as atk
from pygubu.i18n import _
from pygubu.api.v1 import register_widget
from pygubu.plugins.tk.tkstdwidgets import TKLabel
from ..awesometkinter import _plugin_uid, _designer_tabs


class AutoWrappingLabelBO(TKLabel):
    class_ = atk.label.AutoWrappingLabel


_builder_uid = f"{_plugin_uid}.AutoWrappingLabel"
register_widget(
    _builder_uid,
    AutoWrappingLabelBO,
    "AutoWrappingLabel",
    _designer_tabs,
    group=1,
)


class AutofitLabelBO(TKLabel):
    OPTIONS_CUSTOM = ("refresh_time",)
    properties = (
        TKLabel.OPTIONS_STANDARD + TKLabel.OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    )
    ro_properties = TKLabel.ro_properties + OPTIONS_CUSTOM
    class_ = atk.label.AutofitLabel


_builder_uid = f"{_plugin_uid}.AutofitLabel"
register_widget(
    _builder_uid,
    AutofitLabelBO,
    "AutofitLabel",
    _designer_tabs,
    group=1,
)



import os
import logging

from typing import Optional, List
from abc import ABC, abstractmethod


logger = logging.getLogger(__name__)


class IPropertyRegistry(ABC):
    def __init__(self):
        self.properties = {}

    def register(self, name: str, description: dict):
        if name in self.properties:
            self._update_existent(name, description)
        else:
            self._add_new(name, description)

    @abstractmethod
    def _add_new(self, name: str, description: dict):
        ...

    @abstractmethod
    def _update_existent(self: str, name, description: dict):
        ...

    @abstractmethod
    def iter_names(self):
        ...

    @abstractmethod
    def register_custom(
        self,
        builder_uid: str,
        prop_name: str,
        editor: str,
        default_value=None,
        help=None,
        **editor_params,
    ):
        ...


class PropertyRegistryDummy(IPropertyRegistry):
    """If designer is not running there is no need to maintain
    more structures in memory. This class will ignore all method calls.
    """

    def _add_new(self, name: str, description: dict):
        logger.debug("New property %s, ignored.", name)

    def _update_existent(self: str, name, description: dict):
        logger.debug("Update property %s, ignored.", name)

    def iter_names(self):
        return None

    def register_custom(
        self,
        builder_uid: str,
        prop_name: str,
        editor: str,
        default_value=None,
        help=None,
        **editor_params,
    ):
        return None

    def copy_to_builder(
        self, from_builder_id: str, pname: str, to_builder_id: str
    ):
        return None


class PropertyRegistryBase(IPropertyRegistry):
    """Mantain and manage property definitions."""

    def __init__(self):
        super().__init__()
        self._is_sorted: bool = False
        self._sorted_names: List[str] = None

    def _add_new(self, name: str, description: dict):
        self.properties[name] = description
        self._is_sorted = False
        logger.debug("Registered property %s", name)

    def _update_existent(self: str, name, description: dict):
        self.properties[name].update(description)
        logger.debug("Updating registered property %s", name)

    def iter_names(self):
        if not self._is_sorted:
            self._sorted_names = sorted(self.properties.keys())
            self._is_sorted = True
        yield from self._sorted_names

    def register(self, name: str, description: dict):
        if name in self.properties:
            self._update_existent(name, description)
        else:
            self._add_new(name, description)

    def register_custom(
        self,
        builder_uid: str,
        prop_name: str,
        editor: str,
        default_value=None,
        help=None,
        **editor_params,
    ):
        """All custom properties are created using internal dynamic editor."""
        description = {
            "editor": "dynamic",
            builder_uid: {
                "params": {
                    "mode": editor,
                }
            },
        }
        description[builder_uid]["params"].update(editor_params)
        if default_value is not None:
            description[builder_uid]["default"] = default_value
        if help is not None:
            description[builder_uid]["help"] = help
        self.register(prop_name, description)

    def copy_to_builder(
        self, from_builder_id: str, pname: str, to_builder_id: str
    ):
        """Copy property definition from one builder to another."""
        if pname not in self.properties:
            raise RuntimeError(f"Property {pname} not registered.")
        elif from_builder_id not in self.properties[pname]:
            raise RuntimeError(f"Builder ID {from_builder_id} not registered.")
        else:
            description = self.properties[pname][from_builder_id].copy()
            self.properties[pname][to_builder_id] = description


if "PYGUBU_DESIGNER_RUNNING" in os.environ:
    from blinker import Signal

    class PropertyRegistryWithSignals(PropertyRegistryBase):
        on_property_new = Signal()
        # on_property_update = signal()

        def _add_new(self, name: str, description: dict):
            super()._add_new(name, description)
            if self.on_property_new.receivers:
                self.on_property_new.send(name=name)

    PropertyRegistry = PropertyRegistryWithSignals()

else:
    PropertyRegistry = PropertyRegistryDummy()



class TkmtWidgetCodeMixin:
    def code_configure(self, targetid=None):
        return []

    def code_connect_commands(self):
        return []

    def _code_process_pos_kw_properties(self, code_identifier):
        """Creates dict with properties values for code"""
        defaults = self._get_property_defaults()
        args = {}
        for rop in self.properties:
            value_is_set = False
            pvalue = None
            if rop in self.wmeta.properties:
                pvalue = self.wmeta.properties[rop]
                value_is_set = True
            elif rop in defaults:
                pvalue = defaults[rop]
                value_is_set = True
            if value_is_set:
                pvalue = self._code_process_property_value(
                    code_identifier, rop, pvalue
                )
                args[rop] = pvalue
        return args

    def code_realize(self, boparent, code_identifier=None):
        if code_identifier is not None:
            self._code_identifier = code_identifier
        lines = []
        master = boparent.code_child_master()

        pbag = self._code_process_pos_kw_properties(code_identifier)
        kargs = self._get_keyword_args(pbag)
        args = self._get_positional_args(pbag)

        pos_args = ""
        if args:
            pos_args = ",".join(args)

        bag = []
        for pname, value in kargs.items():
            bag.append(f"{pname}={value}")

        kwargs = ""
        if bag:
            if args:
                kwargs = f""", {", ".join(bag)}"""
            else:
                kwargs = ", ".join(bag)
        s = f"{self.code_identifier()} = {self._code_class_name()}({pos_args}{kwargs})"
        if hasattr(self, "master_add_method"):
            method = self.master_add_method
            s = f"{self.code_identifier()} = {master}.{method}({pos_args}{kwargs})"
        lines.append(s)
        return lines



import tkinter as tk
import pygubu.plugins.tk.tkstdwidgets as tkw
import pygubu.plugins.ttk.ttkstdwidgets as ttkw
import pygubu.forms.tkwidget as tkwidget

from pygubu.api.v1 import (
    BuilderObject,
    register_widget,
    register_custom_property,
)
from pygubu.i18n import _
from pygubu.plugins.pygubu.tkscrollbarhelper_bo import TKSBHelperBO
from .base import (
    WidgetBOMixin,
    _plugin_forms_uid,
    _tab_form_widgets_label,
)


_plugin_uid = f"{_plugin_forms_uid}.tkwidget"
_designer_tabs = ("tk", "ttk", _tab_form_widgets_label)


class TextBO(WidgetBOMixin, tkw.TKText):
    class_ = tkwidget.Text
    properties = tkw.TKText.properties + WidgetBOMixin.base_properties
    ro_properties = tkw.TKText.ro_properties + WidgetBOMixin.base_properties


_builder_uid = f"{_plugin_uid}.Text"
_tk_text_builder_uid = _builder_uid

register_widget(
    _builder_uid,
    TextBO,
    "Text",
    _designer_tabs,
    group=0,
)


TKSBHelperBO.add_allowed_child(_builder_uid)



# encoding: utf-8
import os
import sys
import unittest
import tkinter as tk
import tkinter.ttk as ttk

import fixpath
import pygubu
import support


class TestButton(unittest.TestCase):
    def setUp(self):
        support.root_deiconify()
        xmldata = """<?xml version="1.0" ?>
<interface>
  <object class="ttk.Frame" id="mainwindow">
    <property name="height">250</property>
    <property name="width">250</property>
    <layout>
      <property name="column">0</property>
      <property name="sticky">nsew</property>
      <property name="propagate">True</property>
      <property name="row">0</property>
    </layout>
    <child>
      <object class="ttk.Button" id="testbutton">
        <property name="class_">CustomButton</property>
        <property name="command">on_button_click</property>
        <property name="compound">right</property>
        <property name="style">CustomButton.TButton</property>
        <property name="text">Button Label</property>
        <property name="textvariable">button_var</property>
        <layout>
          <property name="column">0</property>
          <property name="propagate">True</property>
          <property name="row">0</property>
        </layout>
      </object>
    </child>
  </object>
</interface>
"""
        self.builder = builder = pygubu.Builder()
        builder.add_from_string(xmldata)
        self.widget = builder.get_object("testbutton")

        self.is_style_setup = False
        if self.is_style_setup:
            self.is_style_setup = True
            s = ttk.Style()
            s.configure("CustomButton.TButton", color="Blue")

    def tearDown(self):
        support.root_withdraw()

    def test_class(self):
        self.assertIsInstance(self.widget, ttk.Button)
        self.widget.destroy()

    def test_class_(self):
        tclobj = self.widget.cget("class")
        class_ = str(tclobj)
        self.assertEqual("CustomButton", class_)
        self.widget.destroy()

    def test_style(self):
        tclobj = self.widget.cget("style")
        style = str(tclobj)
        self.assertEqual("CustomButton.TButton", style)
        self.widget.destroy()

    def test_command_dict(self):
        success = []

        def on_button_click():
            success.append(1)

        cbdic = {"on_button_click": on_button_click}
        self.builder.connect_callbacks(cbdic)

        self.widget.invoke()
        self.assertTrue(success)
        self.widget.destroy()

    def test_command_self(self):
        success = []

        class AnObject:
            def on_button_click(self):
                success.append(1)

        cbobj = AnObject()
        self.builder.connect_callbacks(cbobj)

        self.widget.invoke()
        self.assertTrue(success)
        self.widget.destroy()

    def test_compound(self):
        compound = str(self.widget.cget("compound"))
        self.assertEqual("right", compound)
        self.widget.destroy()

    def test_btn_text(self):
        txt = self.widget.cget("text")
        self.assertEqual("Button Label", txt)
        self.widget.destroy()

    def test_btn_variable(self):
        var = self.builder.get_variable("button_var")
        self.assertIsInstance(var, tk.StringVar)
        self.assertEqual("Button Label", var.get())

        newlabel = "Label Changed"
        var.set(newlabel)
        self.assertEqual(newlabel, self.widget.cget("text"))
        self.widget.destroy()


if __name__ == "__main__":
    unittest.main()



import tkinter as tk
from pygubu.i18n import _
from pygubu.api.v1 import register_widget, register_custom_property
from pygubu.plugins.ttk.ttkstdwidgets import TTKEntry, TTKFrame, TTKCombobox
from ttkwidgets.autocomplete import (
    AutocompleteEntry,
    AutocompleteEntryListbox,
    AutocompleteCombobox,
)

from ..ttkwidgets import _designer_tab_label, _plugin_uid
from .utils import AutocompleteBaseBO


class AutocompleteEntryBO(AutocompleteBaseBO, TTKEntry):
    class_ = AutocompleteEntry
    #
    # Note: completevalues property is managed by AutocompleteBaseBO
    #
    OPTIONS_CUSTOM = ("completevalues",)
    properties = (
        TTKEntry.OPTIONS_STANDARD + TTKEntry.OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    )
    ro_properties = TTKEntry.ro_properties + OPTIONS_CUSTOM

    def code_imports(self):
        return (("ttkwidgets.autocomplete", "AutocompleteEntry"),)


_builder_uid = f"{_plugin_uid}.AutocompleteEntry"

register_widget(
    _builder_uid,
    AutocompleteEntryBO,
    "AutocompleteEntry",
    ("ttk", _designer_tab_label),
    group=4,
)

register_custom_property(
    _builder_uid,
    "completevalues",
    "entry",
    help=_("Values separated by space. In code you can pass a full list"),
)


class AutocompleteEntryListboxBO(AutocompleteBaseBO, TTKFrame):
    class_ = AutocompleteEntryListbox
    container = False
    #
    # Note: completevalues property is managed by AutocompleteBaseBO
    #
    OPTIONS_CUSTOM = (
        "allow_other_values",
        "autohidescrollbar",
        "exportselection",
        "font",
        "justify",
        "completevalues",
    )
    properties = (
        TTKFrame.OPTIONS_STANDARD + TTKFrame.OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    )
    ro_properties = TTKFrame.ro_properties + ("completevalues",)

    def _process_property_value(self, pname, value):
        final_value = value
        if pname in (
            "autohidescrollbar",
            "allow_other_values",
            "exportselection",
        ):
            final_value = tk.getboolean(value)
        else:
            final_value = super(
                AutocompleteEntryListboxBO, self
            )._process_property_value(pname, value)
        return final_value

    def _code_process_property_value(self, targetid, pname, value):
        if pname in (
            "autohidescrollbar",
            "allow_other_values",
            "exportselection",
        ):
            return tk.getboolean(value)
        return super()._code_process_property_value(targetid, pname, value)


_builder_uid = f"{_plugin_uid}.AutocompleteEntryListbox"

register_widget(
    _builder_uid,
    AutocompleteEntryListboxBO,
    "AutocompleteEntryListbox",
    ("ttk", _designer_tab_label),
    group=4,
)

register_custom_property(
    _builder_uid,
    "autohidescrollbar",
    "choice",
    values=("", "true", "false"),
    state="readonly",
)

register_custom_property(
    _builder_uid, "font", "fontentry", help=_("font in entry and listbox")
)
register_custom_property(
    _builder_uid,
    "allow_other_values",
    "choice",
    values=("", "true", "false"),
    state="readonly",
    help=_("whether the user is allowed to enter values not in the list"),
)
register_custom_property(
    _builder_uid,
    "allow_other_values",
    "choice",
    values=("", "true", "false"),
    state="readonly",
    help=_("whether to automatically export selected text to the clipboard"),
)
register_custom_property(
    _builder_uid,
    "justify",
    "choice",
    values=("", "left", "center", "right"),
    state="readonly",
    help=_("text alignment in entry and listbox"),
)
register_custom_property(
    _builder_uid,
    "completevalues",
    "entry",
    help=_(
        "Values separated by space. In code you can pass a full python list."
    ),
)


class AutocompleteComboboxBO(AutocompleteBaseBO, TTKCombobox):
    class_ = AutocompleteCombobox
    OPTIONS_SPECIFIC = tuple(
        set(TTKCombobox.OPTIONS_SPECIFIC) - set(("values",))
    )
    OPTIONS_CUSTOM = ("completevalues",)
    properties = (
        TTKCombobox.OPTIONS_STANDARD + OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    )
    ro_properties = TTKCombobox.ro_properties + OPTIONS_CUSTOM


_builder_uid = f"{_plugin_uid}.AutocompleteCombobox"
register_widget(
    _builder_uid,
    AutocompleteComboboxBO,
    "AutocompleteCombobox",
    ("ttk", _designer_tab_label),
    group=4,
)

register_custom_property(
    _builder_uid,
    "completevalues",
    "entry",
    help=_("Values separated by space. In code you can pass a full list"),
)



# encoding: utf-8
import json
import tkinter as tk
import tkinter.ttk as ttk
from collections import OrderedDict


class Combobox(ttk.Combobox):
    """
    A ttk.Combobox that accepts a list of (key, label) pair of options.
    """

    def __init__(self, *args, keyvariable=None, textvariable=None, **kw):
        self._keyvarcb = None
        self._txtvarcb = None
        self._choices = OrderedDict()
        self._keyvar = tk.StringVar() if keyvariable is None else keyvariable
        self._txtvar = tk.StringVar() if textvariable is None else textvariable

        key = "values"
        if key in kw:
            values = kw[key]
            self.choices = values
            values = self.__choices2tkvalues(self.choices)
            kw[key] = values

        # only normal/disabled supported
        # normal is converted to readonly
        key = "state"
        state = kw.get(key, "readonly")
        if state not in ("readonly", "disabled"):
            state = "readonly"
        kw[key] = state
        kw["textvariable"] = self._txtvar

        super().__init__(*args, **kw)

        self.__config_keyvar(self._keyvar)
        self.__config_txtvar(self._txtvar)

    def __config_keyvar(self, var):
        if var is None:
            return

        def on_keyvar_changed(varname, elementname, mode):
            nvalue = self._keyvar.get()
            if nvalue in self._choices:
                self.current(nvalue)

        keyvar = self._keyvar
        if self._keyvarcb is not None:
            keyvar.trace_remove("write", self._keyvarcb)
        self._keyvar = var
        self._keyvarcb = var.trace_add("write", on_keyvar_changed)

    def __config_txtvar(self, var):
        keyvar = self._keyvar
        if var is None or keyvar is None:
            return

        def on_txtvar_changed(varname, elementname, mode):
            ckey = self.current()
            if ckey is not None:
                self._keyvar.set(ckey)

        txtvar = self._txtvar
        if txtvar and self._txtvarcb is not None:
            txtvar.trace_remove("write", self._txtvarcb)
        self._txtvar = var
        self._txtvarcb = var.trace_add("write", on_txtvar_changed)

    def _delayed_clear_vars(self):
        # clear variables:
        self._clear_tkvar(self._keyvar)
        self._clear_tkvar(self._txtvar)

    def _clear_tkvar(self, variable):
        if isinstance(variable, tk.StringVar):
            variable.set("")

    def configure(self, cnf=None, **kw):
        if cnf:
            return super().configure(cnf, **kw)
        key = "values"
        if key in kw:
            self.choices = kw[key]
            kw[key] = self.__choices2tkvalues(self.choices)
            # clear vars after setting values
            self.after_idle(self._delayed_clear_vars)
        key = "keyvariable"
        if key in kw:
            value = kw.pop(key)
            self.__config_keyvar(value)
        key = "textvariable"
        if key in kw:
            value = kw[key]
            self.__config_txtvar(value)

        # only readonly and disabled supported
        key = "state"
        if key in kw:
            state = kw.get(key, "readonly")
            if state not in ("readonly", "disabled"):
                state = "readonly"
            kw[key] = state
        return super().configure(cnf, **kw)

    config = configure

    def cget(self, key):
        option = "values"
        if key == option:
            return json.dumps(self.choices)
        option = "keyvariable"
        if key == option:
            return self._keyvar
        option = "textvariable"
        if key == option:
            return self._txtvar
        return super().cget(key)

    def __obj2choices(self, values):
        """
        - json list of key, value pairs:
            Example: [["A", "Option 1 Label"], ["B", "Option 2 Label"]]
        - space separated string with a list of options:
            Example: "Option1 Opt2 Opt3"
            will be converted to a key, value pair of the following form:
            (("Option1", "Option1), ("Opt2", "Opt2"), ("Opt3", "Opt3"))
        - an iterable of key, value pairs
        """
        choices = values
        # choices from string
        if isinstance(values, type("")):
            obj = None
            # try json string
            try:
                obj = json.loads(values)
            except ValueError:
                obj = values
            if isinstance(obj, type([])):
                choices = obj
            else:
                choices = obj.split()
        return choices

    def __choices2tkvalues(self, choices):
        """choices: iterable of key, value pairs"""
        values = []
        for k, v in choices:
            values.append(v)
        return values

    @property
    def choices(self):
        return tuple(self._choices.items())

    @choices.setter
    def choices(self, values):
        values = self.__obj2choices(values)
        self._choices.clear()
        for idx, v in enumerate(values):
            key = value = v
            if not isinstance(v, type("")) and len(v) == 2:
                key, value = v
            self._choices[key] = value

    def current(self, key=None):
        """Get the index of the current selection.
        If key is given, return the index of the key.
        """
        idx = super().current(None)
        index = -1
        ckey = None
        # find the key of the current selection,
        # or if a key was given, its index
        for i, k in enumerate(self._choices.keys()):
            if key is None and idx == i:
                ckey = k
                break
            elif key == k:
                index = i
                break
        if key is None:
            # return current key
            return ckey
        return super().current(index)

    def set(self, key):
        value = self._choices[key]
        return super().set(value)

    def get(self):
        return self.current()


if __name__ == "__main__":

    def show_selection(event):
        cbox = event.widget
        print("values:", cbox.cget("values"))
        print("current:", cbox.current())
        var = cbox.cget("keyvariable")
        print("keyvar: ", var.get())
        var = cbox.cget("textvariable")
        print("textvar: ", var.get())

    root = tk.Tk()
    txtvalues = '[["DNI", "Doc.Nac.Id."], ["LE", "Libreta Enrolamientox"]]'
    c = Combobox(root, values=txtvalues)
    c.bind("<<ComboboxSelected>>", show_selection)
    c.configure(values=txtvalues)
    c.set("LE")
    c.grid()

    values = (("tres", 3), (4, "cuatro"), (6, "seis"))
    c = Combobox(root, values=values)
    c.bind("<<ComboboxSelected>>", show_selection)
    c.set(6)
    c.grid()

    var = tk.StringVar()
    var.set("C")
    e = ttk.Entry(root, textvariable=var)
    e.grid()

    var2 = tk.StringVar()
    e2 = ttk.Entry(root, textvariable=var2)
    e2.grid()

    values = (
        ("", "Seleccione Opción"),
        ("A", "A label"),
        ("B", "B label"),
        ("C", "C label"),
    )
    c = Combobox(root, textvariable=var)
    c.bind("<<ComboboxSelected>>", show_selection)
    c.configure(values=values)
    c.set("")
    c.grid()

    c.configure(keyvariable=var2)
    var2.set("A")

    root.rowconfigure(0, weight=1)
    root.columnconfigure(0, weight=1)
    root.mainloop()



from pygubu.api.v1 import IPluginBase, IDesignerPlugin
from pygubu.utils.widget import crop_widget
from pygubu.stockimage import StockRegistry, StockImageCache, StockImage
from .preview import (
    CTkToplevelPreviewBO,
    CTkPreviewBO,
    CTkFramePreviewBO,
    CTkTabviewForPreviewBO,
    CTkSegmentedButtonForPreviewBO,
)
from ..ctkbase import _plugin_uid


class CTkDesignerPlugin(IDesignerPlugin):
    def get_preview_builder(self, builder_uid: str):
        if builder_uid == f"{_plugin_uid}.CTkToplevel":
            return CTkToplevelPreviewBO
        if builder_uid == f"{_plugin_uid}.CTk":
            return CTkPreviewBO
        if builder_uid == f"{_plugin_uid}.CTkFrame":
            return CTkFramePreviewBO
        if builder_uid == f"{_plugin_uid}.CTkTabview":
            return CTkTabviewForPreviewBO
        if builder_uid == f"{_plugin_uid}.CTkSegmentedButton":
            return CTkSegmentedButtonForPreviewBO
        return None

    def get_toplevel_preview_for(
        self, builder_uid: str, widget_id: str, builder, top_master
    ):
        top = None
        toplevel_uids = ("customtkinter.CTkToplevel", "customtkinter.CTk")
        if builder_uid in toplevel_uids:
            # for a new tk root, create a diferent image cache:
            def on_root_created(root):
                image_cache = StockImageCache(root, StockImage.registry)
                builder.image_cache = image_cache

            builder.on_first_object = on_root_created
            top = builder.get_object(widget_id)
        return top

    def configure_for_preview(self, builder_uid: str, widget):
        """Make widget just display with minimal functionality."""

        if not builder_uid.startswith(f"{_plugin_uid}."):
            return

        #
        # do recursive cropping
        #
        crop_widget(widget, recursive=True)

        #
        # Remove default bindings
        #
        def _no_op(event=None):
            pass

        has_canvas = hasattr(widget, "_canvas")
        widget_canvas = widget._canvas if has_canvas else None

        if builder_uid.endswith(".CTkScrollableFrame"):
            return
        elif builder_uid.endswith(".CTKEntry"):
            seqlist = ("<FocusOut>", "<FocusIn>")
            for seq in seqlist:
                widget_canvas.bind(seq, _no_op)
        elif builder_uid.endswith(".CTkSlider"):
            seqlist = ("<Enter>", "<Leave>", "<Button-1>", "<B1-Motion>")
            for seq in seqlist:
                widget_canvas.bind(seq, _no_op)
        elif builder_uid.endswith(".CTkOptionMenu"):
            seqlist = ("<Enter>", "<Leave>", "<Button-1>")
            for seq in seqlist:
                widget_canvas.bind(seq, _no_op)
                widget._text_label.bind(seq, _no_op)
        elif builder_uid.endswith(".CTkComboBox"):
            widget_canvas.tag_bind("right_parts", "<Enter>", _no_op)
            widget_canvas.tag_bind("dropdown_arrow", "<Enter>", _no_op)
            widget_canvas.tag_bind("right_parts", "<Leave>", _no_op)
            widget_canvas.tag_bind("dropdown_arrow", "<Leave>", _no_op)
            widget_canvas.tag_bind("right_parts", "<Button-1>", _no_op)
            widget_canvas.tag_bind("dropdown_arrow", "<Button-1>", _no_op)

    def ensure_visibility_in_preview(self, builder, selected_uid: str):
        """Ensure visibility of selected_uid in preview."""
        xpath = ".//object[@class='customtkinter.CTkTabview.Tab']"
        # find all tabs
        tabs = builder.uidefinition.root.findall(xpath)
        if tabs is None:
            return

        for tab in tabs:
            tab_id = tab.get("id")
            activate_tab = False
            # Check if this tab was clicked
            if tab_id == selected_uid:
                activate_tab = True
            else:
                # check if selected_uid is inside this tab
                xpath = f".//object[@id='{selected_uid}']"
                o = tab.find(xpath)
                if o is not None:
                    activate_tab = True
            if activate_tab:
                tab_builder = builder.objects[tab_id]
                top = tab_builder.widget.winfo_toplevel()
                tabview = top.nametowidget(tab_builder.widget.winfo_parent())
                tabname = tab_builder.wmeta.properties.get("label")
                current = tabview.get()
                if current != tabname:
                    tabview.set(tabname)
                    top.update()
                break



from pygubu.api.v1 import (
    BuilderObject,
    register_widget,
)
from pygubu.plugins.pygubu import _tab_widgets_label, _plugin_uid
from pygubu.widgets.simpletooltip import Tooltip, Tooltipttk
from pygubu.component.builderobject import FamilyType


class TooltipBaseBO(BuilderObject):
    allow_bindings = False
    layout_required = False
    family = FamilyType.MODIFIER

    def realize(self, parent, extra_init_args: dict = None):
        self.widget = parent.get_child_master()
        self.widget.simple_tooltip = self.class_(self.widget)
        return self.widget

    def _set_property(self, target_widget, pname, value):
        target_widget.simple_tooltip.label_options[pname] = value


class SimpleTooltipBO(TooltipBaseBO):
    class_ = Tooltip
    properties = (
        "text",
        "font",
        "background",
        "foreground",
        "justify",
        "wraplength",
        "relief",
        "borderwidth",
        "padx",
        "pady",
    )


_builder_uid = f"{_plugin_uid}.Tooltip"
register_widget(
    _builder_uid,
    SimpleTooltipBO,
    "Tooltip",
    (_tab_widgets_label, "tk", "ttk"),
)


class SimpleTooltipttkBO(SimpleTooltipBO):
    class_ = Tooltipttk
    properties = (
        "text",
        "font",
        "foreground",
        "background",
        "justify",
        "wraplength",
        "relief",
        "borderwidth",
        "style",
    )


_builder_uid = f"{_plugin_uid}.Tooltipttk"
register_widget(
    _builder_uid,
    SimpleTooltipttkBO,
    "Tooltipttk",
    (_tab_widgets_label, "ttk"),
)



from pygubu.api.v1 import register_widget
from pygubu.plugins.ttk.ttkstdwidgets import TTKFrame
from pygubu.widgets.hideableframe import HideableFrame
from pygubu.plugins.pygubu import _tab_widgets_label, _plugin_uid


class HideableFrameBO(TTKFrame):
    class_ = HideableFrame


_builder_uid = f"{_plugin_uid}.hideableframe"
register_widget(
    _builder_uid,
    HideableFrameBO,
    "HideableFrame",
    (_tab_widgets_label, "ttk"),
)



#!/usr/bin/python3
import tkinter as tk
import tkinter.ttk as ttk
from pygubu.api.v1 import (
    BuilderObject,
    register_widget,
)
from pygubu.widgets.colorinput import ColorInput
from pygubu.plugins.pygubu import _tab_widgets_label, _plugin_uid


#
# Builder definition section
#


class ColorInputBO(BuilderObject):
    class_ = ColorInput
    virtual_events = (ColorInput.EVENT_COLOR_CHANGED,)
    properties = ("value", "textvariable")


_builder_id = f"{_plugin_uid}.ColorInput"
register_widget(
    _builder_id, ColorInputBO, "ColorInput", ("ttk", _tab_widgets_label)
)



from pygubu.api.v1 import register_widget
from pygubu.plugins.ttk.ttkstdwidgets import TTKTreeviewBO, TTKTreeviewColumnBO
from pygubu.plugins.pygubu.scrollbarhelper_bo import TTKSBHelperBO
from ttkwidgets.checkboxtreeview import CheckboxTreeview

from ..ttkwidgets import _designer_tab_label, _plugin_uid


class CheckboxTreeviewBO(TTKTreeviewBO):
    class_ = CheckboxTreeview
    allowed_children = ("ttk.Treeview.Column",)


_builder_uid = f"{_plugin_uid}.CheckboxTreeview"
register_widget(
    _builder_uid,
    CheckboxTreeviewBO,
    "CheckboxTreeview",
    ("ttk", _designer_tab_label),
)

TTKSBHelperBO.add_allowed_child(_builder_uid)
TTKTreeviewColumnBO.add_allowed_parent(_builder_uid)



# encoding: utf-8
import tkinter as tk
import tkinter.ttk as ttk
from collections import OrderedDict

from pygubu.i18n import _
from pygubu.api.v1 import BuilderObject, register_widget
from pygubu.component.builderobject import (
    EntryBaseBO,
    PanedWindowBO,
    PanedWindowPaneBO,
    OptionMenuBaseMixin,
    CB_TYPES,
)


has_tk_version_9 = tk.TkVersion >= 9


#
# ttk widgets
#
class TTKWidgetBO(BuilderObject):
    OPTIONS_LABEL = (
        "compound",
        "font",
        "image",
        "padding",
        "text",
        "textvariable",
        "underline",
        "width",
    )
    OPTIONS_COMPATIBILITY = ("state",)
    OPTIONS_STANDARD = ("class_", "cursor", "takefocus", "style")
    OPTIONS_SPECIFIC = tuple()
    OPTIONS_CUSTOM = tuple()
    ro_properties = ("class_",)


class TTKFrame(TTKWidgetBO):
    OPTIONS_SPECIFIC = ("borderwidth", "relief", "padding", "height", "width")
    class_ = ttk.Frame
    container = True
    container_layout = True
    properties = TTKWidgetBO.OPTIONS_STANDARD + OPTIONS_SPECIFIC


register_widget("ttk.Frame", TTKFrame, "ttk.Frame", (_("Containers"), "ttk"))


class TTKLabel(TTKWidgetBO):
    OPTIONS_STANDARD = (
        TTKWidgetBO.OPTIONS_STANDARD
        + TTKWidgetBO.OPTIONS_LABEL
        + ("borderwidth",)
    )
    OPTIONS_SPECIFIC = (
        "anchor",
        "background",
        "font",
        "foreground",
        "justify",
        "padding",
        "relief",
        "state",
        "wraplength",
    )
    class_ = ttk.Label
    container = False
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC


register_widget(
    "ttk.Label", TTKLabel, "ttk.Label", (_("Control & Display"), "ttk")
)


class TTKButton(TTKWidgetBO):
    OPTIONS_STANDARD = (
        TTKWidgetBO.OPTIONS_STANDARD
        + tuple(set(TTKWidgetBO.OPTIONS_LABEL) - set(("font",)))
        + TTKWidgetBO.OPTIONS_COMPATIBILITY
    )
    OPTIONS_SPECIFIC = ("command", "default")
    class_ = ttk.Button
    container = False
    properties = (
        OPTIONS_STANDARD + OPTIONS_SPECIFIC + TTKWidgetBO.OPTIONS_CUSTOM
    )
    command_properties = ("command",)


register_widget(
    "ttk.Button", TTKButton, "ttk.Button", (_("Control & Display"), "ttk")
)


class TTKCheckbutton(TTKWidgetBO):
    OPTIONS_STANDARD = (
        TTKWidgetBO.OPTIONS_STANDARD
        + TTKWidgetBO.OPTIONS_LABEL
        + TTKWidgetBO.OPTIONS_COMPATIBILITY
    )
    OPTIONS_SPECIFIC = ("command", "offvalue", "onvalue", "variable")
    class_ = ttk.Checkbutton
    container = False
    properties = (
        OPTIONS_STANDARD + OPTIONS_SPECIFIC + TTKWidgetBO.OPTIONS_CUSTOM
    )
    command_properties = ("command",)


register_widget(
    "ttk.Checkbutton",
    TTKCheckbutton,
    "ttk.Checkbutton",
    (_("Control & Display"), "ttk"),
)


class TTKRadiobutton(TTKWidgetBO):
    OPTIONS_STANDARD = (
        TTKWidgetBO.OPTIONS_STANDARD
        + TTKWidgetBO.OPTIONS_LABEL
        + TTKWidgetBO.OPTIONS_COMPATIBILITY
    )
    OPTIONS_SPECIFIC = ("command", "value", "variable")
    class_ = ttk.Radiobutton
    container = False
    properties = (
        OPTIONS_STANDARD + OPTIONS_SPECIFIC + TTKWidgetBO.OPTIONS_CUSTOM
    )
    ro_properties = ("class_",)
    command_properties = ("command",)


register_widget(
    "ttk.Radiobutton",
    TTKRadiobutton,
    "ttk.Radiobutton",
    (_("Control & Display"), "ttk"),
)


_v9_entry_opts = (
    ("placeholder", "placeholderforeground") if has_tk_version_9 else tuple()
)


class TTKEntry(TTKWidgetBO, EntryBaseBO):
    OPTIONS_STANDARD = TTKWidgetBO.OPTIONS_STANDARD + ("xscrollcommand",)
    OPTIONS_SPECIFIC = (
        "exportselection",
        "font",
        "invalidcommand",
        "justify",
        "show",
        "state",
        "textvariable",
        "validate",
        "validatecommand",
        "width",
    ) + _v9_entry_opts
    OPTIONS_CUSTOM = ("text",)
    class_ = ttk.Entry
    container = False
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    command_properties = ("validatecommand", "invalidcommand", "xscrollcommand")


register_widget(
    "ttk.Entry", TTKEntry, "ttk.Entry", (_("Control & Display"), "ttk")
)


class TTKCombobox(TTKWidgetBO):
    OPTIONS_SPECIFIC = (
        "exportselection",
        "justify",
        "height",
        "postcommand",
        "state",
        "textvariable",
        "values",
        "width",
        "validate",
        "validatecommand",
        "invalidcommand",
        "xscrollcommand",
    ) + _v9_entry_opts
    class_ = ttk.Combobox
    container = False
    properties = (
        TTKWidgetBO.OPTIONS_STANDARD
        + OPTIONS_SPECIFIC
        + TTKWidgetBO.OPTIONS_CUSTOM
    )
    command_properties = (
        "postcommand",
        "validatecommand",
        "invalidcommand",
        "xscrollcommand",
    )
    virtual_events = ("<<ComboboxSelected>>",)

    def _code_process_property_value(self, targetid, pname, value: str):
        if pname == "values":
            return self.code_escape_str(value)
        return super()._code_process_property_value(targetid, pname, value)


register_widget(
    "ttk.Combobox", TTKCombobox, "ttk.Combobox", (_("Control & Display"), "ttk")
)


class TTKScrollbar(TTKWidgetBO):
    OPTIONS_SPECIFIC = ("command", "orient")
    class_ = ttk.Scrollbar
    container = False
    properties = TTKWidgetBO.OPTIONS_STANDARD + OPTIONS_SPECIFIC
    command_properties = ("command",)


register_widget(
    "ttk.Scrollbar",
    TTKScrollbar,
    "ttk.Scrollbar",
    (_("Control & Display"), "ttk"),
)


class TTKSizegrip(TTKWidgetBO):
    class_ = ttk.Sizegrip
    container = False
    properties = TTKWidgetBO.OPTIONS_STANDARD + TTKWidgetBO.OPTIONS_SPECIFIC


register_widget(
    "ttk.Sizegrip", TTKSizegrip, "ttk.Sizegrip", (_("Control & Display"), "ttk")
)


_v9_pbar_options = (
    ("anchor", "font", "foreground", "justify", "text", "wraplength")
    if has_tk_version_9
    else tuple()
)


class TTKProgressbar(TTKWidgetBO):
    OPTIONS_SPECIFIC = (
        "orient",
        "length",
        "mode",
        "maximum",
        "value",
        "variable",
    ) + _v9_pbar_options
    class_ = ttk.Progressbar
    container = False
    properties = TTKWidgetBO.OPTIONS_STANDARD + OPTIONS_SPECIFIC


register_widget(
    "ttk.Progressbar",
    TTKProgressbar,
    "ttk.Progressbar",
    (_("Control & Display"), "ttk"),
)


class TTKScale(TTKWidgetBO):
    OPTIONS_SPECIFIC = (
        "command",
        "from_",
        "length",
        "orient",
        "state",
        "to",
        "value",
        "variable",
    )
    class_ = ttk.Scale
    container = False
    properties = TTKWidgetBO.OPTIONS_STANDARD + OPTIONS_SPECIFIC
    command_properties = ("command",)


register_widget(
    "ttk.Scale", TTKScale, "ttk.Scale", (_("Control & Display"), "ttk")
)


class TTKSeparator(TTKWidgetBO):
    OPTIONS_SPECIFIC = ("orient",)
    class_ = ttk.Separator
    container = False
    properties = TTKWidgetBO.OPTIONS_STANDARD + OPTIONS_SPECIFIC


register_widget(
    "ttk.Separator",
    TTKSeparator,
    "ttk.Separator",
    (_("Control & Display"), "ttk"),
)


class TTKLabelframe(TTKWidgetBO):
    OPTIONS_STANDARD = TTKFrame.OPTIONS_STANDARD
    OPTIONS_SPECIFIC = TTKFrame.OPTIONS_SPECIFIC + (
        "labelanchor",
        "text",
        "underline",
    )
    class_ = ttk.Labelframe
    container = True
    container_layout = True
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC


register_widget(
    "ttk.Labelframe", TTKLabelframe, "ttk.Labelframe", (_("Containers"), "ttk")
)


class TTKPanedwindow(TTKWidgetBO, PanedWindowBO):
    OPTIONS_SPECIFIC = ("orient", "height", "width")
    class_ = ttk.Panedwindow
    allowed_children = ("ttk.Panedwindow.Pane",)
    properties = TTKWidgetBO.OPTIONS_STANDARD + OPTIONS_SPECIFIC
    ro_properties = ("class_", "orient")
    virtual_events = ("<<EnteredChild>>",)


register_widget(
    "ttk.Panedwindow",
    TTKPanedwindow,
    "ttk.Panedwindow",
    (_("Containers"), "ttk"),
)


class TTKNotebook(TTKWidgetBO):
    OPTIONS_SPECIFIC = ("height", "padding", "width")
    class_ = ttk.Notebook
    container = True
    allowed_children = ("ttk.Notebook.Tab",)
    properties = TTKWidgetBO.OPTIONS_STANDARD + OPTIONS_SPECIFIC
    virtual_events = ("<<NotebookTabChanged>>",)


register_widget(
    "ttk.Notebook", TTKNotebook, "ttk.Notebook", (_("Containers"), "ttk")
)


class TTKMenubuttonBO(TTKWidgetBO):
    OPTIONS_STANDARD = (
        TTKWidgetBO.OPTIONS_STANDARD
        + TTKWidgetBO.OPTIONS_LABEL
        + TTKWidgetBO.OPTIONS_COMPATIBILITY
    )
    OPTIONS_SPECIFIC = ("direction",)  # 'menu'
    class_ = ttk.Menubutton
    container = True
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC
    allowed_children = ("tk.Menu",)
    maxchildren = 1

    def add_child(self, bobject):
        self.widget.configure(menu=bobject.widget)

    def code_child_add(self, childid):
        lines = [f"{self.code_identifier()}.configure(menu={childid})"]
        return lines


register_widget(
    "ttk.Menubutton",
    TTKMenubuttonBO,
    "ttk.Menubutton",
    (
        _("Menu"),
        _("Control & Display"),
        "ttk",
    ),
)


class TTKTreeviewBO(TTKWidgetBO):
    OPTIONS_STANDARD = TTKWidgetBO.OPTIONS_STANDARD + (
        "xscrollcommand",
        "yscrollcommand",
    )
    OPTIONS_SPECIFIC = ("height", "padding", "selectmode", "show")
    class_ = ttk.Treeview
    container = True
    allowed_children = ("ttk.Treeview.Column",)
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC
    command_properties = ("xscrollcommand", "yscrollcommand")
    virtual_events = (
        "<<TreeviewSelect>>",
        "<<TreeviewOpen>>",
        "<<TreeviewClose>>",
    )

    def __init__(self, builder, wdescr):
        super(TTKTreeviewBO, self).__init__(builder, wdescr)
        self._columns = None
        self._headings = None
        self._dcolumns = None

    def configure_children(self):
        self.__configure_columns()

    def __configure_columns(self):
        if self._columns:
            columns = list(self._columns.keys())
            if "#0" in columns:
                columns.remove("#0")
            displaycolumns = self._dcolumns
            self.widget.configure(
                columns=columns, displaycolumns=displaycolumns
            )
            for col in self._columns:
                self.widget.column(col, **self._columns[col])
        if self._headings:
            for col in self._headings:
                self.widget.heading(col, **self._headings[col])

    def set_column(self, col_id, attrs, visible=True):
        if self._columns is None:
            self._columns = OrderedDict()
            self._dcolumns = list()
        self._columns[col_id] = attrs
        if visible and col_id != "#0":
            self._dcolumns.append(col_id)

    def set_heading(self, col_id, attrs):
        if self._headings is None:
            self._headings = OrderedDict()
        self._headings[col_id] = attrs

    #
    # Code generation methods
    #
    def code_configure_children(self, targetid=None):
        if targetid is None:
            targetid = self.code_identifier()
        lines = []
        if self._columns:
            columns = list(self._columns.keys())
            if "#0" in columns:
                columns.remove("#0")
            displaycolumns = self._dcolumns
            line = f"{targetid}_cols = {repr(columns)}"
            lines.append(line)
            line = f"{targetid}_dcols = {repr(displaycolumns)}"
            lines.append(line)
            line = "{0}.configure(columns={0}_cols, displaycolumns={0}_dcols)"
            line = line.format(targetid)
            lines.append(line)
            for col in self._columns:
                code_bag, kwp, _ = self._code_process_properties(
                    self._columns[col], targetid
                )
                bag = []
                for pname in kwp:
                    s = f"{pname}={code_bag[pname]}"
                    bag.append(s)
                kwargs = ",".join(bag)
                line = f'{targetid}.column("{col}", {kwargs})'
                lines.append(line)
        if self._headings:
            for col in self._headings:
                code_bag, kwp, _ = self._code_process_properties(
                    self._headings[col], targetid
                )
                bag = []
                for pname in kwp:
                    s = f"{pname}={code_bag[pname]}"
                    bag.append(s)
                kwargs = ",".join(bag)
                line = f'{targetid}.heading("{col}", {kwargs})'
                lines.append(line)

        return lines


register_widget(
    "ttk.Treeview",
    TTKTreeviewBO,
    "ttk.Treeview",
    (_("Control & Display"), "ttk"),
)


#
# Helpers for Standard ttk widgets
#


class TTKPanedwindowPane(TTKWidgetBO, PanedWindowPaneBO):
    OPTIONS_STANDARD = tuple()
    OPTIONS_SPECIFIC = ("weight",)
    class_ = None
    container = True
    allowed_parents = ("ttk.Panedwindow",)
    maxchildren = 1
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC


register_widget(
    "ttk.Panedwindow.Pane",
    TTKPanedwindowPane,
    "ttk.Panedwindow.Pane",
    (_("Pygubu Helpers"), "ttk"),
)


class TTKNotebookTab(TTKWidgetBO):
    OPTIONS_STANDARD = tuple()
    OPTIONS_SPECIFIC = (
        "state",
        "sticky",
        "padding",
        "text",
        "image",
        "compound",
        "underline",
    )
    class_ = None
    container = True
    layout_required = False
    allow_bindings = False
    allowed_parents = ("ttk.Notebook",)
    children_layout_override = True
    maxchildren = 1
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC

    def realize(self, parent, extra_init_args: dict = None):
        self.widget = parent.get_child_master()
        return self.widget

    def configure(self, target=None):
        pass

    def layout(self, target=None, *, forget=False):
        pass

    def add_child(self, bobject):
        properties = {
            pname: self._process_property_value(pname, value)
            for (pname, value) in self.wmeta.properties.items()
        }
        self.widget.add(bobject.widget, **properties)

    #
    # Code generation methods
    #
    def code_realize(self, boparent, code_identifier=None):
        self._code_identifier = boparent.code_child_master()
        return tuple()

    def code_configure(self, targetid=None):
        return tuple()

    def code_child_add(self, childid):
        targetid = self.code_identifier()
        code_bag, kw, _ = self._code_process_properties(
            self.wmeta.properties, targetid
        )
        kwbag = []
        for pname in kw:
            arg = f"{pname}={code_bag[pname]}"
            kwbag.append(arg)
        kwargs = ""
        if kwbag:
            kwargs = f", {', '.join(kwbag)}"
        line = f"{targetid}.add({childid}{kwargs})"
        return [line]


register_widget(
    "ttk.Notebook.Tab",
    TTKNotebookTab,
    "Notebook.Tab",
    (_("Pygubu Helpers"), "ttk"),
)


class TTKTreeviewColumnBO(TTKWidgetBO):
    OPTIONS_STANDARD = tuple()
    OPTIONS_SPECIFIC = (
        "text",
        "image",
        "command",
        "heading_anchor",
        "column_anchor",
        "minwidth",
        "stretch",
        "width",
    )
    OPTIONS_CUSTOM = (
        "tree_column",
        "visible",
    )
    class_ = None
    container = False
    layout_required = False
    allow_bindings = False
    allowed_parents = ("ttk.Treeview",)
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    command_properties = ("command",)

    def realize(self, parent, extra_init_args: dict = None):
        self.widget = parent.get_child_master()
        col_props = dict(self.wmeta.properties)  # copy properties
        self._setup_column(parent, col_props)
        return self.widget

    def _get_heading_properties(self, props):
        text = props.pop("text", None)
        if text is None:
            text = self.wmeta.identifier
        hprops = {"anchor": props.pop("heading_anchor", tk.W), "text": text}
        # Only add image if has value. Fix code generation
        imgvalue = props.pop("image", None)
        if imgvalue:
            hprops["image"] = self._process_property_value("image", imgvalue)
        return hprops

    def _get_column_properties(self, props):
        cprops = {
            "anchor": props.pop("column_anchor", ""),
            "stretch": props.pop("stretch", "1"),
            "width": props.pop("width", "200"),
            "minwidth": props.pop("minwidth", "20"),
        }
        return cprops

    def _setup_column(self, parent, col_props):
        tree_column = col_props.pop("tree_column", "false")
        tree_column = tree_column.lower()
        tree_column = True if tree_column == "true" else False
        column_id = "#0" if tree_column else self.wmeta.identifier
        visible = col_props.pop("visible", "true")
        visible = visible.lower()
        is_visible = True if visible == "true" else False

        # configure heading properties
        col_props.pop("command", "")
        hprops = self._get_heading_properties(col_props)
        parent.set_heading(column_id, hprops)

        # configure column properties
        cprops = self._get_column_properties(col_props)
        parent.set_column(column_id, cprops, is_visible)

    def configure(self, target=None):
        pass

    def layout(self, target=None, *, forget=False):
        pass

    def _connect_command(self, cpname, callback):
        tree_column = self.wmeta.properties.get("tree_column", "false")
        tree_column = tree_column.lower()
        tree_column = True if tree_column == "true" else False
        column_id = "#0" if tree_column else self.wmeta.identifier
        self.widget.heading(column_id, command=callback)

    #
    # Code generation methods
    #
    def code_realize(self, boparent, code_identifier=None):
        self.parent_bo = boparent
        self._code_identifier = boparent.code_child_master()
        col_props = dict(self.wmeta.properties)  # copy properties
        self._setup_column(boparent, col_props)
        return tuple()

    def code_configure(self, targetid=None):
        return tuple()

    def _code_connect_command(self, cmd_pname, cmd, cbname):
        target = self.parent_bo.code_identifier()

        args = cmd.get("args", "")
        args = args.split() if args else None
        lines = []
        cmdtype = cmd["cbtype"]
        if cmdtype == CB_TYPES.WITH_WID:
            wid = self.wmeta.identifier
            fname = f"{wid}_cmd_"
            fdef = f"""def {fname}(): {cbname}("{wid}")\n"""
            cbname = fname
            lines.append(fdef)

        tree_column = self.wmeta.properties.get("tree_column", "false")
        tree_column = True if tree_column.lower() == "true" else False
        column_id = "#0" if tree_column else self.wmeta.identifier
        lines.append(
            f"""{target}.heading("{column_id}", {cmd_pname}={cbname})"""
        )
        return lines


register_widget(
    "ttk.Treeview.Column",
    TTKTreeviewColumnBO,
    "Treeview.Column",
    (_("Pygubu Helpers"), "ttk"),
)


class TTKSpinboxBO(TTKWidgetBO, EntryBaseBO):
    OPTIONS_STANDARD = TTKEntry.OPTIONS_STANDARD
    OPTIONS_SPECIFIC = (
        TTKEntry.OPTIONS_SPECIFIC
        + (
            "from_",
            "to",
            "increment",
            "values",
            "wrap",
            "format",
            "command",
        )
        + _v9_entry_opts
    )
    OPTIONS_CUSTOM = TTKEntry.OPTIONS_CUSTOM
    class_ = None
    container = False
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    command_properties = (
        "validatecommand",
        "invalidcommand",
        "xscrollcommand",
        "command",
    )
    virtual_events = ("<<Increment>>", "<<Decrement>>")


if tk.TkVersion >= 8.5:
    # Note:
    # ttk::Spinbox was added in tk 8.5.9 so it may fail in lower 8.5 patch versions

    if not hasattr(ttk, "Spinbox"):
        from pygubu.widgets.ttkspinbox import Spinbox

        ttk.Spinbox = Spinbox

    TTKSpinboxBO.class_ = ttk.Spinbox

    register_widget(
        "ttk.Spinbox",
        TTKSpinboxBO,
        "ttk.Spinbox",
        (_("Control & Display"), "ttk"),
    )


class OptionMenuBO(OptionMenuBaseMixin, BuilderObject):
    class_ = ttk.OptionMenu
    properties = (
        "style",
        "direction",
        "command",
        "variable",
        "value",
        "values",
    )
    command_properties = ("command",)
    ro_properties = ("variable", "value", "values")

    def _create_option_menu(self, master, variable, value, values, command):
        return self.class_(master, variable, value, *values, command=command)

    def _code_create_optionmenu(
        self,
        identifier,
        classname,
        master,
        value_arg,
        variable_arg,
        command_arg,
    ):
        return (
            f"{identifier} = {classname}({master}, {variable_arg},"
            + f" {value_arg}, *__values, command={command_arg})"
        )


register_widget(
    "ttk.OptionMenu",
    OptionMenuBO,
    "ttk.OptionMenu",
    (_("Control & Display"), "ttk"),
)


class LabeledScaleBO(BuilderObject):
    class_ = ttk.LabeledScale
    properties = (
        "compound",
        "variable",
        "from_",
        "to",
    )
    ro_properties = ("compound", "from_", "to", "variable")
    virtual_events = ("<<RangeChanged>>",)

    def _connect_binding(self, sequence: str, callback, add):
        self.widget.scale.bind(sequence, callback, add)

    def _code_connect_binding(
        self, target: str, sequence: str, callback: str, add_arg: str
    ):
        scale = f"{target}.scale"
        return super()._code_connect_binding(scale, sequence, callback, "+")


register_widget(
    "ttk.LabeledScale",
    LabeledScaleBO,
    "ttk.LabeledScale",
    (_("Control & Display"), "ttk"),
)



import logging
import math
import tkinter as tk
from .exceptions import ImageFormatNotSupportedError, StockImageException
from .registry import StockRegistry, StockItem
from .config import _img_notsupported, _img_notfound


logger = logging.getLogger(__name__)


class StockImageCache:
    IMAGE_NOT_SUPPORTED = "img_not_supported"
    IMAGE_NOT_FOUND = "img_not_found"
    BASELINE = 1.33398982438864281

    def __init__(
        self, tkroot, registry: StockRegistry, *, auto_scaling: bool = False
    ):
        self.tkroot = tkroot
        self.registry: StockRegistry = registry
        self._cached = {}
        if not registry.is_registered(self.IMAGE_NOT_SUPPORTED):
            self.register_from_data(
                self.IMAGE_NOT_SUPPORTED, "gif", _img_notsupported
            )
        if not registry.is_registered(self.IMAGE_NOT_FOUND):
            self.register_from_data(self.IMAGE_NOT_FOUND, "gif", _img_notfound)
        self.auto_scaling = auto_scaling
        self._scale_factor = None

    def clear_cache(self):
        """Call this before closing tk root"""
        # Prevent tkinter errors on python 2 ??
        for key in self._cached:
            self._cached[key] = None
        self._cached = {}

    def register(self, image_id, filename):
        """Register a image file using image_id"""
        self.registry.register(image_id, filename)

    def register_from_package(self, image_id, fpath):
        self.registry.register_from_package(image_id, fpath)

    def register_from_data(self, image_id, format, data):
        """Register a image data using image_id"""
        self.registry.register_from_data(image_id, format, data)

    def register_created(self, image_id, image):
        """Register an already created image using image_id"""

        if image_id in self._cached:
            logger.warning("Warning, replacing resource {0}", image_id)
        self._cached[image_id] = image
        logger.info("data registered as %s", image_id)

    def is_registered(self, image_id):
        return image_id in self._cached or self.registry.is_registered(image_id)

    def register_all_from_dir(
        self, dir_path, prefix=None, ext=None, recurse=False, fullname_key=False
    ):
        self.registry.register_all_from_dir(
            dir_path, prefix, ext, recurse, fullname_key
        )

    def register_all_from_pkg(
        self, pkg, prefix=None, ext=None, recurse=False, fullname_key=False
    ):
        self.registry.register_all_from_pkg(
            pkg, prefix, ext, recurse, fullname_key
        )

    def get(self, image_id, custom_loader=None):
        """Get image previously registered with key image_id.
        If key not exist, raise StockImageException
        """

        if image_id in self._cached:
            logger.info("Resource %s is in cache.", image_id)
            return self._cached[image_id]
        if self.registry.is_registered(image_id):
            img = self._load_image(image_id, custom_loader)
            return img
        else:
            raise StockImageException(f"StockImage: {image_id} not registered.")

    def _setup_root(self):
        if self.tkroot is None:
            self.tkroot = tk._get_default_root()

    def _load_image(self, image_id, custom_loader=None):
        """Load image from file or return the cached instance."""

        stock_item: StockItem = self.registry.get_item(image_id)
        img = None
        try:
            img = stock_item.create_image(
                tk_master=self.tkroot, custom_loader=custom_loader
            )
            if isinstance(img, tk.PhotoImage):
                img = self._process_photo_image(img)
        except ImageFormatNotSupportedError:
            msg = "Error loading image %s, try installing Pillow module."
            logger.error(msg, image_id)
            img = self.get(self.IMAGE_NOT_SUPPORTED)

        self._cached[image_id] = img
        logger.info("Loaded resource data for %s.", image_id)
        return img

    @property
    def scale_factor(self):
        if self._scale_factor is None:
            self._setup_root()
            scaling = self.tkroot.tk.call("tk", "scaling")
            self._scale_factor = scaling / self.BASELINE
        return self._scale_factor

    def _process_photo_image(self, image: tk.PhotoImage):
        """Additional processing for image before adding to cache."""
        new_image = image
        if self.auto_scaling:
            new_image = self._resize_photo(image, self.scale_factor)
        return new_image

    def _resize_photo(self, img: tk.PhotoImage, scale_factor: float):
        """Resize image using tk.PhotoImage methods.
        scale_factor is rounded to the nearest 0.5 value.
        Resizing starts if scale factor is greater or equal to 1.5,
        otherwise image is returned with no changes.
        """
        w = img.width()
        new_w = math.ceil(w * scale_factor)
        # h = img.height()
        # new_h = math.ceil(h * scale_factor)

        # round(number / roundto) * roundto
        xfactor = round((new_w / w) / 0.5) * 0.5
        if xfactor < 1.5:
            # No scaling for xfactor below 1.5
            return img
        xfactor_par = xfactor % 2 == 0.0
        if xfactor_par:
            # zoom scaling
            zoom_factor = int(xfactor)
            new_img = img.zoom(zoom_factor)
        else:
            # zoom + subsample scaling
            zoom_factor = int(xfactor * 2)
            zimg = img.zoom(zoom_factor)
            new_img = zimg.subsample(2)
        return new_img

    def as_iconbitmap(self, image_id):
        """Get image path for use in iconbitmap property"""
        return self.registry.as_iconbitmap(image_id)

    def add_resource_path(self, path):
        self.registry.add_resource_path(path)

    def add_resource_package(self, package):
        self.registry.add_resource_package(package)

    def find_and_register(self, image_id):
        self.registry.find_and_register(image_id)



from pygubu.widgets.combobox import Combobox
from pygubu.widgets.fontinput import FontInput

from .ttkwidget import TtkVarBasedWidget


class PygubuCombobox(TtkVarBasedWidget, Combobox):
    tkvar_pname = "keyvariable"


class FontInputFW(TtkVarBasedWidget, FontInput):
    """A FontInput form field widget."""

    tkvar_pname = "variable"



import logging
from ...component.builderobject import (
    CLASS_MAP,
    WidgetDescription,
)
from ...component.property_registry import PropertyRegistry


logger = logging.getLogger(__name__)


def register_builder(
    classname, builder, label=None, tags=None, group=None, public=True
):
    if label is None:
        label = classname
    if tags is None:
        tags = tuple()
    if group is None:
        group = 9

    CLASS_MAP[classname] = WidgetDescription(
        classname, builder, label, tags, group, public
    )


def register_widget(
    classname, builder, label=None, tags=None, group=None, public=True
):
    return register_builder(classname, builder, label, tags, group, public)


def register_property(name, description):
    PropertyRegistry.register(name, description)


def register_custom_property(
    builder_uid,
    prop_name,
    editor,
    default_value=None,
    help=None,
    **editor_params,
):
    """Helper function to register a custom property.
    All custom properties are created using internal dynamic editor.
    """
    PropertyRegistry.register_custom(
        builder_uid, prop_name, editor, default_value, help, **editor_params
    )


def copy_custom_property(from_builder_id, pname, to_builder_id):
    PropertyRegistry.copy_to_builder(from_builder_id, pname, to_builder_id)



# encoding: utf-8
from pygubu.api.v1 import register_widget
from pygubu.plugins.ttk.ttkstdwidgets import (
    TTKTreeviewBO,
    TTKTreeviewColumnBO,
)
from .scrollbarhelper_bo import TTKSBHelperBO
from pygubu.widgets.editabletreeview import EditableTreeview
from pygubu.plugins.pygubu import _tab_widgets_label, _plugin_uid


class EditableTreeviewBO(TTKTreeviewBO):
    class_ = EditableTreeview
    virtual_events = TTKTreeviewBO.virtual_events + (
        "<<TreeviewInplaceEdit>>",
        "<<TreeviewCellEdited>>",
        "<<TreeviewEditorsUnfocused>>",
    )


_builder_uid = f"{_plugin_uid}.EditableTreeview"
if _builder_uid not in TTKTreeviewColumnBO.allowed_parents:
    TTKTreeviewColumnBO.add_allowed_parent(_builder_uid)
if _builder_uid not in TTKSBHelperBO.allowed_children:
    TTKSBHelperBO.add_allowed_child(_builder_uid)

register_widget(
    _builder_uid,
    EditableTreeviewBO,
    "EditableTreeview",
    (_tab_widgets_label, "ttk"),
)

# Register old name until removal
_old_name = "pygubu.builder.widgets.editabletreeview"
register_widget(_old_name, EditableTreeviewBO, public=False)
if _old_name not in TTKTreeviewColumnBO.allowed_parents:
    TTKTreeviewColumnBO.add_allowed_parent(_old_name)
if _old_name not in TTKSBHelperBO.allowed_children:
    TTKSBHelperBO.add_allowed_child(_old_name)



import importlib
from pygubu.i18n import _
from pygubu.api.v1 import BuilderLoaderPlugin


_designer_tab_label = _("TkinterModernThemes")
_plugin_uid = "tkmt"


class TkmthemesLoader(BuilderLoaderPlugin):
    module_map = {
        "pygubu.plugins.tkmt.widgets": (f"{_plugin_uid}.ThemedTKinterFrame",),
    }

    def do_activate(self) -> bool:
        spec = importlib.util.find_spec("TKinterModernThemes")
        return spec is not None

    def get_module_for(self, identifier: str) -> str:
        for module, identifiers in self.module_map.items():
            if identifier in identifiers:
                return module
        return None

    def get_all_modules(self):
        return [m for m in self.module_map.keys()]

    def can_load(self, identifier: str) -> bool:
        for module, builders in self.module_map.items():
            if identifier in builders:
                return True
        return False

    def get_designer_plugin(self):
        """Load class that implements IDesignerPlugin"""
        from .designer.designerplugin import PygubuDesignerPlugin

        return PygubuDesignerPlugin()



import awesometkinter as atk
from pygubu.api.v1 import (
    BuilderObject,
    register_widget,
)

from ..awesometkinter import _plugin_uid, _designer_tabs


class RadialProgressbarBO(BuilderObject):
    OPTIONS_CUSTOM = (
        "bg",
        "fg",
        "text_fg",
        "font",
        "font_size_ratio",
        "base_img",
        "indicator_img",
        "parent_bg",
    )
    ro_properties = OPTIONS_CUSTOM
    class_ = atk.RadialProgressbar

    def _process_property_value(self, pname, value):
        if pname == "font_size_ratio":
            value_ = 0.1
            try:
                value_ = float(value)
            except ValueError:
                pass
            return value_
        return super()._process_property_value(pname, value)


_builder_uid = f"{_plugin_uid}.RadialProgressbar"
register_widget(
    _builder_uid,
    RadialProgressbarBO,
    "RadialProgressbar",
    _designer_tabs,
)


class RadialProgressbar3dBO(BuilderObject):
    OPTIONS_CUSTOM = ("fg", "text_fg", "text_bg")
    ro_properties = OPTIONS_CUSTOM
    class_ = atk.RadialProgressbar3d


_builder_uid = f"{_plugin_uid}.RadialProgressbar3d"
register_widget(
    _builder_uid,
    RadialProgressbar3dBO,
    "RadialProgressbar3d",
    _designer_tabs,
)


class SegmentbarBO(BuilderObject):
    OPTIONS_CUSTOM = ("bg", "fg", "width", "height")
    ro_properties = OPTIONS_CUSTOM
    class_ = atk.Segmentbar

    def _process_property_value(self, pname, value):
        if pname in ("width",):
            return int(value)
        return super()._process_property_value(pname, value)


_builder_uid = f"{_plugin_uid}.Segmentbar"
register_widget(_builder_uid, SegmentbarBO, "Segmentbar", _designer_tabs)



import tkinter as tk

from pygubu.api.v1 import BuilderObject, register_widget
from pygubu.plugins.ttk.ttkstdwidgets import TTKNotebookTab
from tkinterweb.utilities import Notebook, ColourSelector
from ..tkinterweb import _designer_tab_label, _plugin_uid


class NotebookBO(BuilderObject):
    class_ = Notebook
    container = True
    container_layout = True
    ro_properties = ("class_", "takefocus")
    properties = (
        # ttk.Frame properties:
        "class_",
        "cursor",
        "style",
        "borderwidth",
        "relief",
        "padding",
        "height",
        "width",
    )
    virtual_events = ("<<NotebookTabChanged>>",)


_builder_uid = f"{_plugin_uid}.Notebook"
_notebook_uid = _builder_uid
register_widget(
    _builder_uid, NotebookBO, "Notebook", ("ttk", _designer_tab_label), group=1
)


class NotebookPageBO(TTKNotebookTab):
    allowed_parents = (_notebook_uid,)
    children_layout_override = True
    properties = (
        "state",
        "text",
        "image",
        "compound",
        "underline",
    )


_builder_uid = f"{_plugin_uid}.Notebook.Page"
NotebookBO.add_allowed_child(_builder_uid)

register_widget(
    _builder_uid,
    NotebookPageBO,
    "Notebook.Page",
    ("ttk", _designer_tab_label),
    group=1,
)


class ColourSelectorBO(BuilderObject):
    class_ = ColourSelector
    properties = ("colour",)
    ro_properties = ("colour",)
    virtual_events = ("<<Modified>>",)


_builder_uid = f"{_plugin_uid}.ColourSelector"
register_widget(
    _builder_uid,
    ColourSelectorBO,
    "ColourSelector",
    ("ttk", _designer_tab_label),
    group=2,
)



from pygubu.i18n import _
from pygubu.api.v1 import register_widget, register_custom_property
from pygubu.plugins.ttk.ttkstdwidgets import TTKFrame
from ttkwidgets import Calendar

from ..ttkwidgets import _designer_tab_label, _plugin_uid


class CalendarBO(TTKFrame):
    class_ = Calendar
    container = False
    OPTIONS_CUSTOM = (
        "locale",
        "firstweekday",
        "year",
        "month",
        "selectbackground",
        "selectforeground",
    )
    properties = (
        TTKFrame.OPTIONS_STANDARD + TTKFrame.OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    )
    # FIXME: set all custom properties as readonly, there is a bug in the
    # ttkwidgets widget definition.
    ro_properties = TTKFrame.ro_properties + OPTIONS_CUSTOM

    def _process_property_value(self, pname, value):
        if pname in ("firstweekday", "year", "month"):
            return int(value)
        return super()._process_property_value(pname, value)


_builder_uid = f"{_plugin_uid}.Calendar"
register_widget(
    _builder_uid, CalendarBO, "Calendar", ("ttk", _designer_tab_label), group=2
)

register_custom_property(
    _builder_uid,
    "locale",
    "entry",
    help=_("calendar locale (defines the language, date formatting)"),
)
register_custom_property(
    _builder_uid,
    "firstweekday",
    "choice",
    values=("", 0, 1, 2, 3, 4, 5, 6),
    state="readonly",
    help=_("first day of the week, 0 is monday"),
)
register_custom_property(
    _builder_uid, "year", "integernumber", help=_("year to display")
)
register_custom_property(
    _builder_uid,
    "month",
    "choice",
    values=("", 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12),
    state="readonly",
    help=_("month to display"),
)
register_custom_property(
    _builder_uid,
    "selectbackground",
    "colorentry",
    help=_("background color of the selected day"),
)
register_custom_property(
    _builder_uid,
    "selectforeground",
    "colorentry",
    help=_("selectforeground color of the selected day"),
)



# encoding: utf-8
import os
import sys
import unittest
import tkinter as tk
import tkinter.ttk as ttk


import fixpath
import pygubu
import support


class TestTreeview(unittest.TestCase):
    def setUp(self):
        support.root_deiconify()
        xmldata = """<?xml version="1.0" ?>
<interface>
  <object class="ttk.Frame" id="mainwindow">
    <property name="height">250</property>
    <property name="width">250</property>
    <layout>
      <property name="sticky">nesw</property>
      <property name="row">0</property>
      <property name="column">0</property>
      <property name="propagate">True</property>
    </layout>
    <child>
      <object class="ttk.Treeview" id="treeview">
        <property name="selectmode">browse</property>
        <layout>
          <property name="column">0</property>
          <property name="row">0</property>
          <property name="propagate">True</property>
        </layout>
        <child>
          <object class="ttk.Treeview.Column" id="treecolumn">
            <property name="tree_column">True</property>
            <property name="visible">True</property>
            <property name="text">Tree</property>
            <property name="command">on_treecolumn_click</property>
            <property name="heading_anchor">w</property>
            <property name="column_anchor">w</property>
            <property name="minwidth">200</property>
            <property name="stretch">True</property>
            <property name="width">200</property>
          </object>
        </child>
        <child>
          <object class="ttk.Treeview.Column" id="column1">
            <property name="tree_column">False</property>
            <property name="visible">True</property>
            <property name="text">Column 1</property>
            <property name="heading_anchor">center</property>
            <property name="column_anchor">w</property>
            <property name="minwidth">200</property>
            <property name="stretch">False</property>
            <property name="width">200</property>
          </object>
        </child>
        <child>
          <object class="ttk.Treeview.Column" id="hidden_column">
            <property name="tree_column">False</property>
            <property name="visible">False</property>
            <property name="text">hidden</property>
            <property name="heading_anchor">w</property>
            <property name="column_anchor">w</property>
            <property name="minwidth">20</property>
            <property name="stretch">True</property>
            <property name="width">200</property>
          </object>
        </child>
      </object>
    </child>
  </object>
</interface>
"""
        self.builder = builder = pygubu.Builder()
        builder.add_from_string(xmldata)
        self.widget = builder.get_object("treeview")
        self.widget.wait_visibility()

    def tearDown(self):
        support.root_withdraw()

    def test_class(self):
        self.assertIsInstance(self.widget, ttk.Treeview)
        self.widget.destroy()

    def test_selectmode(self):
        expected = "browse"
        value = str(self.widget.cget("selectmode"))
        self.assertEqual(expected, value)
        self.widget.destroy()

    def test_columns(self):
        columns = ("column1", "hidden_column")
        wcolumns = self.widget.cget("columns")
        self.assertEqual(columns, wcolumns)

        dcolumns = ("column1",)
        wdcolumns = self.widget.cget("displaycolumns")
        self.assertEqual(dcolumns, wdcolumns)

        self.widget.destroy()

    def test_tree_heading(self):
        wh = self.widget.heading("#0")
        heading = {
            "text": "Tree",
            "anchor": "w",
        }
        for k, v in heading.items():
            self.assertEqual(v, wh[k])
        self.widget.destroy()

    def test_tree_column(self):
        wc = self.widget.column("#0")
        column = {
            "anchor": "w",
            "stretch": 1,
            "width": 200,
            "minwidth": 200,
        }

        for k, v in column.items():
            self.assertEqual(v, wc[k])
        self.widget.destroy()

    def test_command_dict(self):
        success = []

        def on_treecolumn_click():
            success.append(1)

        cbdic = {"on_treecolumn_click": on_treecolumn_click}
        self.builder.connect_callbacks(cbdic)

        x, y = self.widget.winfo_x(), self.widget.winfo_y()
        self.widget.event_generate("<ButtonPress-1>", x=x + 5, y=y + 5)
        self.widget.event_generate("<ButtonRelease-1>", x=x + 5, y=y + 5)
        self.widget.update()

        self.assertTrue(success)
        self.widget.destroy()

    def test_command_self(self):
        success = []

        class AnObject:
            def on_treecolumn_click(self):
                success.append(1)

        cbobj = AnObject()
        self.builder.connect_callbacks(cbobj)

        x, y = self.widget.winfo_x(), self.widget.winfo_y()
        self.widget.event_generate("<ButtonPress-1>", x=x + 5, y=y + 5)
        self.widget.event_generate("<ButtonRelease-1>", x=x + 5, y=y + 5)
        self.widget.update()

        self.assertTrue(success)
        self.widget.destroy()


if __name__ == "__main__":
    unittest.main()



from pygubu.api.v1 import register_widget, register_custom_property
from pygubu.i18n import _
from pygubu.plugins.tk.tkstdwidgets import TKToplevel as TKToplevelBO
from pygubu.plugins.tk.tkstdwidgets import TKRootBO
from customtkinter import (
    CTk,
    CTkToplevel,
    set_appearance_mode,
    set_default_color_theme,
)

from ..customtkinter import _designer_tab_label, _plugin_uid
from .ctkbase import CTkBaseMixin


class CTkBO(TKRootBO):
    class_ = CTk
    properties = (
        "cursor",
        "padx",
        "pady",
        "relief",
        "takefocus",
        "height",
        "width",
        # CUSTOM OPTIONS,
        "title",
        "geometry",
        "overrideredirect",
        "minsize",
        "maxsize",
        "resizable",
        "iconbitmap",
        "iconphoto",
        # CUSTOMTKINTER options,
        "appearance_mode",
        "color_theme",
        "fg_color",
    )
    ro_properties = ("container", "fg_color")

    def realize(self, parent, extra_init_args: dict = None):
        args = self._get_init_args(extra_init_args)
        # master = parent.get_child_master()
        fg_color = args.pop("fg_color", None)
        self.widget = self.class_(fg_color, **args)
        return self.widget

    def _set_property(self, target_widget, pname, value):
        if pname == "appearance_mode":
            set_appearance_mode(value)
        elif pname == "color_theme":
            set_default_color_theme(value)
        else:
            return super()._set_property(target_widget, pname, value)

    #
    # Code generation methods
    #

    def code_realize(self, boparent, code_identifier=None):
        if code_identifier is not None:
            self._code_identifier = code_identifier
        lines = []
        # master = boparent.code_child_master()
        init_args = self._code_get_init_args(self.code_identifier())
        fg_color = init_args.pop("fg_color", None)
        bag = []
        for pname, value in init_args.items():
            bag.append(f"{pname}={value}")
        kwargs = ""
        if bag:
            kwargs = f""", {", ".join(bag)}"""
        s = f"{self.code_identifier()} = {self._code_class_name()}({fg_color}{kwargs})"
        lines.append(s)
        return lines

    def _code_set_property(self, targetid, pname, value, code_bag):
        if pname == "appearance_mode":
            code_bag[pname] = (f'set_appearance_mode("{value}")',)
        elif pname == "color_theme":
            code_bag[pname] = (f'set_default_color_theme("{value}")',)
        else:
            super()._code_set_property(targetid, pname, value, code_bag)

    def code_imports(self):
        imports = ["CTk"]
        if "appearance_mode" in self.wmeta.properties:
            imports.append("set_appearance_mode")
        if "color_theme" in self.wmeta.properties:
            imports.append("set_default_color_theme")
        return [("customtkinter", name) for name in imports]


_builder_uid = f"{_plugin_uid}.CTk"
register_widget(
    _builder_uid,
    CTkBO,
    "CTk",
    ("ttk", _designer_tab_label),
    group=-1,
)

_maxsize_help = _("Set the maximum window size.")
_minsize_help = _("Set the minimum window size.")

register_custom_property(
    _builder_uid,
    "minsize",
    "whentry",
    help=_minsize_help,
)
register_custom_property(
    _builder_uid,
    "maxsize",
    "whentry",
    help=_maxsize_help,
)


class CTkToplevelBO(TKToplevelBO):
    class_ = CTkToplevel
    properties = (
        "borderwidth",
        "cursor",
        "takefocus",
        "height",
        "width",
        "highlightbackground",
        "highlightthickness",
        # CUSTOM OPTIONS,
        "title",
        "geometry",
        "overrideredirect",
        "minsize",
        "maxsize",
        "resizable",
        "iconbitmap",
        "iconphoto",
        # CUSTOMTKINTER options,
        "fg_color",
    )
    ro_properties = ("fg_color",)

    def realize(self, parent, extra_init_args: dict = None):
        kwargs = self._get_init_args(extra_init_args)
        fg_color = kwargs.pop("fg_color", None)
        master = parent.get_child_master()
        self.widget = self.class_(master, fg_color=fg_color, **kwargs)
        return self.widget

    def code_imports(self):
        return [("customtkinter", self.class_.__name__)]


_builder_uid = f"{_plugin_uid}.CTkToplevel"
register_widget(
    _builder_uid,
    CTkToplevelBO,
    "CTkToplevel",
    ("ttk", _designer_tab_label),
    group=-1,
)

register_custom_property(
    _builder_uid,
    "minsize",
    "whentry",
    help=_minsize_help,
)
register_custom_property(
    _builder_uid,
    "maxsize",
    "whentry",
    help=_maxsize_help,
)



# encoding: utf-8
import tkinter as tk
import tkinter.ttk as ttk

from pygubu.binding import (
    ApplicationLevelBindManager as BindManager,
    remove_binding,
)
from .tkscrolledframe import ScrolledFrameBase


class ScrolledFrame(ScrolledFrameBase, ttk.Frame):
    _framecls = ttk.Frame
    _sbarcls = ttk.Scrollbar



import os
from pygubu.api.v1 import BuilderLoaderPlugin
from pygubu.i18n import _


_tab_widgets_label = _("Pygubu Widgets")
_tab_helpers_label = _("Pygubu Helpers")
_plugin_uid = "pygubu.widgets"


class PygubuWidgetsLoader(BuilderLoaderPlugin):
    module_map = {
        "pygubu.plugins.pygubu.accordionframe_bo": (
            f"{_plugin_uid}.AccordionFrame",
        ),
        "pygubu.plugins.pygubu.calendarframe_bo": (
            f"{_plugin_uid}.CalendarFrame",
            "pygubu.builder.widgets.calendarframe",
        ),
        "pygubu.plugins.pygubu.colorinput_bo": (f"{_plugin_uid}.ColorInput",),
        "pygubu.plugins.pygubu.combobox_bo": (
            f"{_plugin_uid}.Combobox",
            "pygubu.builder.widgets.combobox",
        ),
        "pygubu.plugins.pygubu.dialog_bo": (
            f"{_plugin_uid}.Dialog",
            "pygubu.builder.widgets.dialog",
        ),
        "pygubu.plugins.pygubu.dockfw_bo": (
            f"{_plugin_uid}.dockframe",
            f"{_plugin_uid}.dockpane",
            f"{_plugin_uid}.dockwidget",
        ),
        "pygubu.plugins.pygubu.editabletreeview_bo": (
            f"{_plugin_uid}.EditableTreeview",
            "pygubu.builder.widgets.editabletreeview",
        ),
        "pygubu.plugins.pygubu.filterabletreeview_bo": (
            f"{_plugin_uid}.FilterableTreeview",
        ),
        "pygubu.plugins.pygubu.floodgauge_bo": (f"{_plugin_uid}.Floodgauge",),
        "pygubu.plugins.pygubu.fontinputbo": (f"{_plugin_uid}.FontInput",),
        # Forms are not finished so expect changes
        "pygubu.plugins.pygubu.forms.pygubuwidget_bo": (
            "pygubu.forms.pygubuwidget.PygubuCombobox"
        ),
        "pygubu.plugins.pygubu.forms.tkwidget_bo": (
            "pygubu.forms.tkwidget.Text",
        ),
        "pygubu.plugins.pygubu.forms.ttkwidget_bo": (
            "pygubu.forms.ttkwidget.FrameFormBuilder",
            "pygubu.forms.ttkwidget.Label",
            "pygubu.forms.ttkwidget.Entry",
            "pygubu.forms.ttkwidget.LabelWidgetInfo",
        ),
        "pygubu.plugins.pygubu.hideableframe_bo": (
            f"{_plugin_uid}.hideableframe",
        ),
        "pygubu.plugins.pygubu.pathchooserinput_bo": (
            "pygubu.builder.widgets.pathchooserinput",
            "pygubu.widgets.PathChooserInput",
            f"{_plugin_uid}.PathChooserButton",
        ),
        "pygubu.plugins.pygubu.scrollbarhelper_bo": (
            "pygubu.builder.widgets.scrollbarhelper",
            f"{_plugin_uid}.ScrollbarHelper",
        ),
        "pygubu.plugins.pygubu.scrolledframe_bo": (
            "pygubu.builder.widgets.scrolledframe",
            f"{_plugin_uid}.ScrolledFrame",
        ),
        "pygubu.plugins.pygubu.simpletooltip_bo": (
            f"{_plugin_uid}.Tooltip",
            f"{_plugin_uid}.Tooltipttk",
        ),
        "pygubu.plugins.pygubu.tkscrollbarhelper_bo": (
            "pygubu.builder.widgets.tkscrollbarhelper",
            f"{_plugin_uid}.TkScrollbarHelper",
        ),
        "pygubu.plugins.pygubu.tkscrolledframe_bo": (
            "pygubu.builder.widgets.tkscrolledframe",
            f"{_plugin_uid}.TkScrolledFrame",
        ),
    }

    def do_activate(self) -> bool:
        return True

    def get_all_modules(self):
        return [m for m in self.module_map.keys()]

    def get_module_for(self, identifier: str) -> str:
        for module, identifiers in self.module_map.items():
            if identifier in identifiers:
                return module
        return None

    def can_load(self, identifier: str) -> bool:
        for module, builders in self.module_map.items():
            if identifier in builders:
                return True
        return False

    def get_designer_plugin(self):
        """Load class that implements IDesignerPlugin"""
        from .designer.designerplugin import PygubuDesignerPlugin

        return PygubuDesignerPlugin()



# encoding: utf-8
import os
import sys
import unittest
import tkinter as tk
import tkinter.ttk as ttk

import fixpath
import support
import pygubu


class TestBuilder01(unittest.TestCase):
    def setUp(self):
        support.root_deiconify()
        xmldata = "test_builder_forget_unnamed.ui"
        self.builder = builder = pygubu.Builder()
        builder.add_from_file(xmldata)
        self.widget = builder.get_object("fmain")
        builder.connect_callbacks({})
        builder.forget_unnamed()

    def tearDown(self):
        support.root_withdraw()

    def test_builder_forget_unnamed(self):
        expected = ["fmain", "mylabel1", "mylabel2"]
        result = list(self.builder.objects.keys())
        self.assertEqual(expected, result)
        self.widget.destroy()



# encoding: utf-8
from pygubu.api.v1 import (
    BuilderObject,
    register_widget,
    register_custom_property,
)
from pygubu.plugins.ttk.ttkstdwidgets import TTKFrame
from pygubu.widgets.calendarframe import CalendarFrame
from pygubu.plugins.pygubu import _tab_widgets_label, _plugin_uid


class CalendarFrameBuilder(BuilderObject):
    class_ = CalendarFrame
    OPTIONS_STANDARD = TTKFrame.OPTIONS_STANDARD
    OPTIONS_SPECIFIC = TTKFrame.OPTIONS_SPECIFIC
    OPTIONS_CUSTOM = (
        "firstweekday",
        "year",
        "month",
        "calendarfg",
        "calendarbg",
        "headerfg",
        "headerbg",
        "selectbg",
        "selectfg",
        "state",
        "markbg",
        "markfg",
        "linewidth",
    )
    ro_properties = TTKFrame.ro_properties
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    virtual_events = ("<<CalendarFrameDateSelected>>",)


_builder_id = f"{_plugin_uid}.CalendarFrame"
register_widget(
    _builder_id,
    CalendarFrameBuilder,
    "CalendarFrame",
    ("ttk", _tab_widgets_label),
)
# Register old name until removal
register_widget(
    "pygubu.builder.widgets.calendarframe", CalendarFrameBuilder, public=False
)



from .form import Form, FormField


class FormBuilder:
    def __init__(self, *args, field_name: str, **kw):
        super().__init__(*args, **kw)
        self.field_name = field_name
        self.fields_scanned = False
        self.widgets = {}
        self.widgets_info = {}
        self.form_info = {}

    def iter_widgets(self, force_scan=False):
        if self.fields_scanned is False or force_scan:
            self.scan_widgets()
        for name, widget in self.widgets.items():
            yield name, widget
        self.fields_scanned = True

    def scan_widgets(self):
        pass

    def get_form(self, config: dict):
        form = Form()
        for name, widget in self.iter_widgets():
            params = {}
            if name in config:
                params = config[name]
            field = FormField(widget, **params)
            form.add(name, field)
        for name, info in self.widgets_info.items():
            form.add_info_display(name, info)
        trans_key = "translator"
        if trans_key in config:
            form.translator = config[trans_key]
        return form



import importlib
from pygubu.i18n import _
from pygubu.api.v1 import BuilderLoaderPlugin

_designer_tab_label = _("ttkwidgets")
_plugin_uid = "ttkwidgets"


class TtkWidgetsLoader(BuilderLoaderPlugin):
    module_map = {
        "pygubu.plugins.ttkwidgets.calendar": (f"{_plugin_uid}.Calendar",),
        "pygubu.plugins.ttkwidgets.checkboxtreeview": (
            f"{_plugin_uid}.CheckboxTreeview",
        ),
        "pygubu.plugins.ttkwidgets.itemscanvas": (
            f"{_plugin_uid}.ItemsCanvas",
        ),
        "pygubu.plugins.ttkwidgets.linklabel": (f"{_plugin_uid}.LinkLabel",),
        "pygubu.plugins.ttkwidgets.scaleentry": (f"{_plugin_uid}.ScaleEntry",),
        "pygubu.plugins.ttkwidgets.scrolledlistbox": (
            f"{_plugin_uid}.ScrolledListbox",
        ),
        "pygubu.plugins.ttkwidgets.table": (f"{_plugin_uid}.Table",),
        "pygubu.plugins.ttkwidgets.tickscale": (f"{_plugin_uid}.TickScale",),
        "pygubu.plugins.ttkwidgets.frames": (
            f"{_plugin_uid}.ScrolledFrame",
            f"{_plugin_uid}.ToggledFrame",
        ),
        "pygubu.plugins.ttkwidgets.color": (
            f"{_plugin_uid}.AlphaBar",
            f"{_plugin_uid}.ColorSquare",
            f"{_plugin_uid}.GradientBar",
        ),
        "pygubu.plugins.ttkwidgets.autocomplete": (
            f"{_plugin_uid}.AutocompleteEntry",
            f"{_plugin_uid}.AutocompleteEntryListbox",
            f"{_plugin_uid}.AutocompleteCombobox",
        ),
        "pygubu.plugins.ttkwidgets.font": (
            f"{_plugin_uid}.FontFamilyDropdown",
            f"{_plugin_uid}.FontFamilyListbox",
            f"{_plugin_uid}.FontSelectFrame",
            f"{_plugin_uid}.FontPropertiesFrame",
            f"{_plugin_uid}.FontSizeDropdown",
        ),
    }

    def do_activate(self) -> bool:
        spec = importlib.util.find_spec("ttkwidgets")
        return spec is not None

    def get_module_for(self, identifier: str) -> str:
        for module, identifiers in self.module_map.items():
            if identifier in identifiers:
                return module
        return None

    def get_all_modules(self):
        return [m for m in self.module_map.keys()]

    def can_load(self, identifier: str) -> bool:
        return identifier.startswith("ttkwidgets.")



import json


class ListDTO:
    def __init__(self, default_value=None, on_error_default=None):
        self.default_value = default_value
        self.on_error_default = (
            ["Invalid data"] if on_error_default is None else on_error_default
        )

    def transform(self, value: str, default=None):
        val = self.default_value if default is None else default
        try:
            json_list = json.loads(value)
            if isinstance(json_list, list):
                val = json_list
            else:
                val = self.on_error_default
        except json.JSONDecodeError:
            val = self.on_error_default
        return val



from abc import ABC, abstractmethod
from typing import Optional, Any
from urllib.parse import urlparse
from functools import reduce


class InvalidURIError(Exception):
    ...


class IDataPool(ABC):
    """A generic pool of data."""

    @abstractmethod
    def get_resource(self, uri: str) -> Any:
        ...

    def get_uri_path(self, uri: str) -> str:
        try:
            result = urlparse(uri)
            if all([result.scheme == "res", result.netloc]):
                return f"{result.netloc}{result.path}"
        except AttributeError:
            pass
        raise InvalidURIError(uri)


class DictDataPool(IDataPool):
    def __init__(self, data: dict = None):
        self.data = {} if data is None else data

    def get_resource(self, uri: str) -> Any:
        path = self.get_uri_path(uri)
        return self.deep_get(self.data, path)

    def deep_get(self, dictionary, keys, default=None):
        return reduce(
            lambda d, key: d.get(key, default)
            if isinstance(d, dict)
            else default,
            keys.split("/"),
            dictionary,
        )


if __name__ == "__main__":
    data = {"value_a": 185, "lists": {"dogs": [10, 20, 30]}}
    pool = DictDataPool(data)
    rlist = [
        "res://value_a",
        "res://lists/dogs",
        "something",
        "res://lists/dogs//failed",
    ]
    for uri in rlist:
        try:
            res = pool.get_resource(uri)
            print(res)
        except InvalidURIError:
            print("Invalid uri error:", uri)



# encoding: utf-8
from pygubu.api.v1 import register_widget
from pygubu.plugins.ttk.ttkstdwidgets import TTKCombobox
from pygubu.widgets.combobox import Combobox
from pygubu.plugins.pygubu import _tab_widgets_label, _plugin_uid


class ComboboxBO(TTKCombobox):
    OPTIONS_SPECIFIC = tuple(
        set(TTKCombobox.OPTIONS_SPECIFIC) - set(("state",))
    )
    OPTIONS_CUSTOM = TTKCombobox.OPTIONS_CUSTOM + ("keyvariable",)
    properties = (
        TTKCombobox.OPTIONS_STANDARD
        + TTKCombobox.OPTIONS_SPECIFIC
        + OPTIONS_CUSTOM
    )
    tkvar_properties = TTKCombobox.tkvar_properties + ("keyvariable",)
    class_ = Combobox


ComboboxBuilder = ComboboxBO  # Alias for future rename of builder.


_builder_id = f"{_plugin_uid}.Combobox"
register_widget(
    _builder_id, ComboboxBO, "Combobox", ("ttk", _tab_widgets_label)
)

# Register old name until removal
register_widget("pygubu.builder.widgets.combobox", ComboboxBO, public=False)



from pygubu.i18n import _
from pygubu.api.v1 import (
    BuilderObject,
    register_widget,
)
from tksheet import Sheet
from ..tksheet import _designer_tab_label, _plugin_uid


class SheetBuilder(BuilderObject):
    class_ = Sheet


_builder_uid = f"{_plugin_uid}.Sheet"
register_widget(
    _builder_uid, SheetBuilder, "Sheet", ("ttk", _designer_tab_label)
)



import tkinter as tk

# from pygubu.i18n import _
from pygubu.api.v1 import register_widget
from tkinterweb import HtmlLabel, HtmlFrame
from pygubu.plugins.ttk.ttkstdwidgets import TTKFrame
from ..tkinterweb import _designer_tab_label, _plugin_uid


class HtmlFrameBO(TTKFrame):
    class_ = HtmlFrame
    container = False
    OPTIONS_CUSTOM = (
        "messages_enabled",
        "vertical_scrollbar",
        "horizontal_scrollbar",
    )
    properties = (
        TTKFrame.OPTIONS_STANDARD + TTKFrame.OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    )
    ro_properties = OPTIONS_CUSTOM

    def _process_property_value(self, pname, value):
        if pname == "messages_enabled":
            return tk.getboolean(value)
        if pname in ("vertical_scrollbar", "horizontal_scrollbar"):
            return value if value == "auto" else tk.getboolean(value)
        return super()._process_property_value(pname, value)

    def _code_process_property_value(self, targetid, pname, value):
        if pname in self.OPTIONS_CUSTOM:
            newval = self._process_property_value(pname, value)
            return newval if isinstance(newval, bool) else f"'{newval}'"
        return super()._code_process_property_value(targetid, pname, value)

    def code_imports(self):
        return (("tkinterweb", "HtmlFrame"),)


_builder_uid = f"{_plugin_uid}.HtmlFrame"
register_widget(
    _builder_uid,
    HtmlFrameBO,
    "HtmlFrame",
    ("ttk", _designer_tab_label),
    group=0,
)


class HtmlLabelBO(HtmlFrameBO):
    class_ = HtmlLabel
    OPTIONS_CUSTOM = ("messages_enabled", "text")
    properties = (
        HtmlFrameBO.OPTIONS_STANDARD
        + HtmlFrameBO.OPTIONS_SPECIFIC
        + OPTIONS_CUSTOM
    )
    ro_properties = OPTIONS_CUSTOM

    def code_imports(self):
        return (("tkinterweb", "HtmlLabel"),)


_builder_uid = f"{_plugin_uid}.HtmlLabel"
register_widget(
    _builder_uid,
    HtmlLabelBO,
    "HtmlLabel",
    ("ttk", _designer_tab_label),
    group=0,
)



import awesometkinter as atk
from pygubu.api.v1 import register_widget
from pygubu.plugins.tk.tkstdwidgets import TKCheckbutton
from pygubu.plugins.ttk.ttkstdwidgets import TTKButton, TTKRadiobutton
from ..awesometkinter import _plugin_uid, _designer_tabs


class Button3dBO(TTKButton):
    OPTIONS_STANDARD = tuple(set(TTKButton.OPTIONS_STANDARD) - set(("style",)))
    OPTIONS_CUSTOM = ("bg", "fg")
    properties = OPTIONS_STANDARD + TTKButton.OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    ro_properties = TTKButton.ro_properties + ("bg", "fg")
    class_ = atk.Button3d


_builder_uid = f"{_plugin_uid}.Button3d"
register_widget(_builder_uid, Button3dBO, "Button3d", _designer_tabs, group=2)


class RadiobuttonBO(TTKRadiobutton):
    OPTIONS_STANDARD = tuple(
        set(TTKRadiobutton.OPTIONS_STANDARD) - set(("style",))
    )
    OPTIONS_CUSTOM = (
        "bg",
        "fg",
        "ind_bg",
        "ind_mark_color",
        "ind_outline_color",
        "font",
    )
    properties = (
        OPTIONS_STANDARD + TTKRadiobutton.OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    )
    ro_properties = (
        TTKRadiobutton.ro_properties + OPTIONS_CUSTOM + ("text", "value")
    )
    class_ = atk.Radiobutton


_builder_uid = f"{_plugin_uid}.Radiobutton"
register_widget(
    _builder_uid,
    RadiobuttonBO,
    "Radiobutton",
    _designer_tabs,
    group=2,
)


class CheckbuttonBO(TKCheckbutton):
    OPTIONS_CUSTOM = (
        "box_color",
        "check_mark_color",
        "text_color",
    )
    properties = (
        TKCheckbutton.OPTIONS_STANDARD
        + TKCheckbutton.OPTIONS_SPECIFIC
        + OPTIONS_CUSTOM
    )
    ro_properties = TKCheckbutton.ro_properties + OPTIONS_CUSTOM
    class_ = atk.Checkbutton


_builder_uid = f"{_plugin_uid}.Checkbutton"
register_widget(
    _builder_uid,
    CheckbuttonBO,
    "Checkbutton",
    _designer_tabs,
    group=2,
)



import importlib
from pygubu.i18n import _
from pygubu.api.v1 import IPluginBase, IBuilderLoaderPlugin, IDesignerPlugin


_designer_tab_label = _("CustomTkinter")
_plugin_uid = "customtkinter"


class CTkBuilderLoader(IBuilderLoaderPlugin, IPluginBase):
    module_map = {
        "pygubu.plugins.customtkinter.windows": (
            f"{_plugin_uid}.CTkToplevel",
            f"{_plugin_uid}.CTk",
        ),
        "pygubu.plugins.customtkinter.widgets": (
            f"{_plugin_uid}.CTkFrame",
            f"{_plugin_uid}.CTkLabel",
            f"{_plugin_uid}.CTkProgressBar",
            f"{_plugin_uid}.CTkButton",
            f"{_plugin_uid}.CTkSlider",
            f"{_plugin_uid}.CTkEntry",
            f"{_plugin_uid}.CTkOptionMenu",
            f"{_plugin_uid}.CTkComboBox",
            f"{_plugin_uid}.CTkCheckBox",
            f"{_plugin_uid}.CTkRadioButton",
            f"{_plugin_uid}.CTkSwitch",
            f"{_plugin_uid}.CTkTextbox",
            f"{_plugin_uid}.CTkCanvas",
            f"{_plugin_uid}.CTkScrollbar",
            f"{_plugin_uid}.CTkTabview",
            f"{_plugin_uid}.CTkScrollableFrame",
        ),
    }

    def do_activate(self) -> bool:
        spec = importlib.util.find_spec("customtkinter")
        return spec is not None

    def get_module_for(self, identifier: str) -> str:
        for module, identifiers in self.module_map.items():
            if identifier in identifiers:
                return module
        return None

    def get_all_modules(self):
        return [m for m in self.module_map.keys()]

    def can_load(self, identifier: str) -> bool:
        return identifier.startswith("customtkinter.")

    def get_designer_plugin(self):
        """Load class that implements IDesignerPlugin"""
        from .designer.designerplugin import CTkDesignerPlugin

        return CTkDesignerPlugin()
